import copy
from functools import partial
from typing import Tuple

import numpy as np
import torch

from hive.agents.dqn import DQNAgent
from hive.agents.qnets.base import FunctionApproximator
from hive.agents.qnets.noisy_linear import NoisyLinear
from hive.agents.qnets.qnet_heads import (
    DistributionalNetwork,
    DQNNetwork,
    DuelingNetwork,
)
from hive.agents.qnets.utils import InitializationFn, calculate_output_dim
from hive.replays import PrioritizedReplayBuffer
from hive.replays.replay_buffer import BaseReplayBuffer
from hive.utils.loggers import Logger
from hive.utils.schedule import Schedule
from hive.utils.utils import LossFn, OptimizerFn, seeder


class RainbowDQNAgent(DQNAgent):
    """An agent implementing the Rainbow algorithm."""

    def __init__(
        self,
        representation_net: FunctionApproximator,
        obs_dim: Tuple,
        act_dim: int,
        optimizer_fn: OptimizerFn = None,
        loss_fn: LossFn = None,
        init_fn: InitializationFn = None,
        id=0,
        replay_buffer: BaseReplayBuffer = None,
        discount_rate: float = 0.99,
        n_step: int = 1,
        grad_clip: float = None,
        reward_clip: float = None,
        update_period_schedule: Schedule = None,
        target_net_soft_update: bool = False,
        target_net_update_fraction: float = 0.05,
        target_net_update_schedule: Schedule = None,
        epsilon_schedule: Schedule = None,
        test_epsilon: float = 0.001,
        min_replay_history: int = 5000,
        batch_size: int = 32,
        device="cpu",
        logger: Logger = None,
        log_frequency: int = 100,
        noisy: bool = True,
        std_init: float = 0.5,
        use_eps_greedy: bool = False,
        double: bool = True,
        dueling: bool = True,
        distributional: bool = True,
        v_min: float = 0,
        v_max: float = 200,
        atoms: int = 51,
    ):
        """
        Args:
            representation_net (FunctionApproximator): A network that outputs the
                representations that will be used to compute Q-values (e.g.
                everything except the final layer of the DQN).
            obs_dim: The shape of the observations.
            act_dim (int): The number of actions available to the agent.
            id: Agent identifier.
            optimizer_fn (OptimizerFn): A function that takes in a list of parameters
                to optimize and returns the optimizer. If None, defaults to
                :py:class:`~torch.optim.Adam`.
            loss_fn (LossFn): Loss function used by the agent. If None, defaults to
                :py:class:`~torch.nn.SmoothL1Loss`.
            init_fn (InitializationFn): Initializes the weights of qnet using
                create_init_weights_fn.
            replay_buffer (BaseReplayBuffer): The replay buffer that the agent will
                push observations to and sample from during learning. If None,
                defaults to
                :py:class:`~hive.replays.prioritized_replay.PrioritizedReplayBuffer`.
            discount_rate (float): A number between 0 and 1 specifying how much
                future rewards are discounted by the agent.
            n_step (int): The horizon used in n-step returns to compute TD(n) targets.
            grad_clip (float): Gradients will be clipped to between
                [-grad_clip, grad_clip].
            reward_clip (float): Rewards will be clipped to between
                [-reward_clip, reward_clip].
            update_period_schedule (Schedule): Schedule determining how frequently
                the agent's Q-network is updated.
            target_net_soft_update (bool): Whether the target net parameters are
                replaced by the qnet parameters completely or using a weighted
                average of the target net parameters and the qnet parameters.
            target_net_update_fraction (float): The weight given to the target
                net parameters in a soft update.
            target_net_update_schedule (Schedule): Schedule determining how frequently
                the target net is updated.
            epsilon_schedule (Schedule): Schedule determining the value of epsilon
                through the course of training.
            test_epsilon (float): epsilon (probability of choosing a random action)
                to be used during testing phase.
            min_replay_history (int): How many observations to fill the replay buffer
                with before starting to learn.
            batch_size (int): The size of the batch sampled from the replay buffer
                during learning.
            device: Device on which all computations should be run.
            logger (ScheduledLogger): Logger used to log agent's metrics.
            log_frequency (int): How often to log the agent's metrics.
            noisy (bool): Whether to use noisy linear layers for exploration.
            std_init (float): The range for the initialization of the standard
                deviation of the weights.
            use_eps_greedy (bool): Whether to use epsilon greedy exploration.
            double (bool): Whether to use double DQN.
            dueling (bool): Whether to use a dueling network architecture.
            distributional (bool): Whether to use the distributional RL.
            vmin (float): The minimum of the support of the categorical value
                distribution for distributional RL.
            vmax (float): The maximum of the support of the categorical value
                distribution for distributional RL.
            atoms (int): Number of atoms discretizing the support range of the
                categorical value distribution for distributional RL.
        """

        self._noisy = noisy
        self._std_init = std_init
        self._double = double
        self._dueling = dueling
        self._distributional = distributional

        self._atoms = atoms if self._distributional else 1
        self._v_min = v_min
        self._v_max = v_max

        if loss_fn is None:
            loss_fn = torch.nn.MSELoss

        if replay_buffer is None:
            replay_buffer = PrioritizedReplayBuffer(seed=seeder.get_new_seed())

        super().__init__(
            representation_net,
            obs_dim,
            act_dim,
            optimizer_fn=optimizer_fn,
            init_fn=init_fn,
            loss_fn=loss_fn,
            id=id,
            replay_buffer=replay_buffer,
            discount_rate=discount_rate,
            n_step=n_step,
            grad_clip=grad_clip,
            reward_clip=reward_clip,
            target_net_soft_update=target_net_soft_update,
            target_net_update_fraction=target_net_update_fraction,
            target_net_update_schedule=target_net_update_schedule,
            update_period_schedule=update_period_schedule,
            epsilon_schedule=epsilon_schedule,
            test_epsilon=test_epsilon,
            min_replay_history=min_replay_history,
            batch_size=batch_size,
            device=device,
            logger=logger,
            log_frequency=log_frequency,
        )

        self._supports = torch.linspace(
            self._v_min, self._v_max, self._atoms, device=self._device
        )

        self._use_eps_greedy = use_eps_greedy

    def create_q_networks(self, representation_net):
        """Creates the Q-network and target Q-network. Adds the appropriate heads
        for DQN, Dueling DQN, Noisy Networks, and Distributional DQN.

        Args:
            representation_net: A network that outputs the representations that will
                be used to compute Q-values (e.g. everything except the final layer
                of the DQN).
        """
        network = representation_net(self._obs_dim)
        network_output_dim = np.prod(calculate_output_dim(network, self._obs_dim))

        # Use NoisyLinear when creating output heads if noisy is true
        linear_fn = (
            partial(NoisyLinear, std_init=self._std_init)
            if self._noisy
            else torch.nn.Linear
        )

        # Set up Dueling heads
        if self._dueling:
            network = DuelingNetwork(
                network, network_output_dim, self._act_dim, linear_fn, self._atoms
            )
        else:
            network = DQNNetwork(
                network, network_output_dim, self._act_dim * self._atoms, linear_fn
            )

        # Set up DistributionalNetwork wrapper if distributional is true
        if self._distributional:
            self._qnet = DistributionalNetwork(
                network, self._act_dim, self._v_min, self._v_max, self._atoms
            )
        else:
            self._qnet = network
        self._qnet.to(device=self._device)
        self._qnet.apply(self._init_fn)
        self._target_qnet = copy.deepcopy(self._qnet).requires_grad_(False)

    @torch.no_grad()
    def act(self, observation):

        if self._training:
            if not self._learn_schedule.get_value():
                epsilon = 1.0
            elif not self._use_eps_greedy:
                epsilon = 0.0
            else:
                epsilon = self._epsilon_schedule.update()
            if self._logger.update_step(self._timescale):
                self._logger.log_scalar("epsilon", epsilon, self._timescale)
        else:
            epsilon = self._test_epsilon

        observation = torch.tensor(
            np.expand_dims(observation, axis=0), device=self._device
        ).float()
        qvals = self._qnet(observation)

        if self._rng.random() < epsilon:
            action = self._rng.integers(self._act_dim)
        else:
            action = torch.argmax(qvals).item()

        if (
            self._training
            and self._logger.should_log(self._timescale)
            and self._state["episode_start"]
        ):
            self._logger.log_scalar("train_qval", torch.max(qvals), self._timescale)
            self._state["episode_start"] = False

        return action

    def update(self, update_info):
        """
        Updates the DQN agent.
        Args:
            update_info: dictionary containing all the necessary information to
            update the agent. Should contain a full transition, with keys for
            "observation", "action", "reward", "next_observation", and "done".
        """
        if update_info["done"]:
            self._state["episode_start"] = True

        if not self._training:
            return

        # Add the most recent transition to the replay buffer.
        self._replay_buffer.add(**self.preprocess_update_info(update_info))

        # Update the q network based on a sample batch from the replay buffer.
        # If the replay buffer doesn't have enough samples, catch the exception
        # and move on.
        if (
            self._learn_schedule.update()
            and self._replay_buffer.size() > 0
            and self._update_period_schedule.update()
        ):
            batch = self._replay_buffer.sample(batch_size=self._batch_size)
            (
                current_state_inputs,
                next_state_inputs,
                batch,
            ) = self.preprocess_update_batch(batch)

            # Compute predicted Q values
            self._optimizer.zero_grad()
            pred_qvals = self._qnet(*current_state_inputs)
            actions = batch["action"].long()

            if self._double:
                next_action = self._qnet(*next_state_inputs)
            else:
                next_action = self._target_qnet(*next_state_inputs)
            next_action = next_action.argmax(1)

            if self._distributional:
                current_dist = self._qnet.dist(*current_state_inputs)
                probs = current_dist[torch.arange(actions.size(0)), actions]
                probs = torch.clamp(probs, 1e-6, 1)  # NaN-guard
                log_p = torch.log(probs)
                with torch.no_grad():
                    target_prob = self.target_projection(
                        next_state_inputs, next_action, batch["reward"], batch["done"]
                    )

                loss = -(target_prob * log_p).sum(-1)

            else:
                pred_qvals = pred_qvals[torch.arange(pred_qvals.size(0)), actions]

                next_qvals = self._target_qnet(*next_state_inputs)
                next_qvals = next_qvals[torch.arange(next_qvals.size(0)), next_action]

                q_targets = batch["reward"] + self._discount_rate * next_qvals * (
                    1 - batch["done"]
                )

                loss = self._loss_fn(pred_qvals, q_targets)

            if isinstance(self._replay_buffer, PrioritizedReplayBuffer):
                td_errors = loss.sqrt().detach().cpu().numpy()
                self._replay_buffer.update_priorities(batch["indices"], td_errors)
                loss *= batch["weights"]
            loss = loss.mean()

            if self._logger.should_log(self._timescale):
                self._logger.log_scalar(
                    "train_loss",
                    loss,
                    self._timescale,
                )
            loss.backward()
            if self._grad_clip is not None:
                torch.nn.utils.clip_grad_value_(
                    self._qnet.parameters(), self._grad_clip
                )
            self._optimizer.step()

        # Update target network
        if self._target_net_update_schedule.update():
            self._update_target()

    def target_projection(self, target_net_inputs, next_action, reward, done):
        """Project distribution of target Q-values.

        Args:
            target_net_inputs: Inputs to feed into the target net to compute the
                projection of the target Q-values. Should be set from
                :py:meth:`~hive.agents.dqn.DQNAgent.preprocess_update_batch`.
            next_action (~torch.Tensor): Tensor containing next actions used to
                compute target distribution.
            reward (~torch.Tensor): Tensor containing rewards for the current batch.
            done (~torch.Tensor): Tensor containing whether the states in the current
                batch are terminal.

        """
        reward = reward.reshape(-1, 1)
        not_done = 1 - done.reshape(-1, 1)
        batch_size = reward.size(0)
        next_dist = self._target_qnet.dist(*target_net_inputs)
        next_dist = next_dist[torch.arange(batch_size), next_action]

        dist_supports = reward + not_done * self._discount_rate * self._supports
        dist_supports = dist_supports.clamp(min=self._v_min, max=self._v_max)
        dist_supports = dist_supports.unsqueeze(1)
        dist_supports = dist_supports.tile([1, self._atoms, 1])
        projected_supports = self._supports.tile([batch_size, 1]).unsqueeze(2)

        delta = float(self._v_max - self._v_min) / (self._atoms - 1)
        quotient = 1 - (torch.abs(dist_supports - projected_supports) / delta)
        quotient = quotient.clamp(min=0, max=1)

        projection = torch.sum(quotient * next_dist.unsqueeze(1), dim=2)
        return projection



import gym

from hive.envs.base import BaseEnv
from hive.envs.env_spec import EnvSpec


class GymEnv(BaseEnv):
    """
    Class for loading gym environments.
    """

    def __init__(self, env_name, num_players=1, **kwargs):
        """
        Args:
            env_name (str): Name of the environment (NOTE: make sure it is available
                in gym.envs.registry.all())
            num_players (int): Number of players for the environment.
            kwargs: Any arguments you want to pass to :py:meth:`create_env` or
                :py:meth:`create_env_spec` can be passed as keyword arguments to this
                constructor.
        """
        self.create_env(env_name, **kwargs)
        super().__init__(self.create_env_spec(env_name, **kwargs), num_players)

    def create_env(self, env_name, **kwargs):
        """Function used to create the environment. Subclasses can override this method
        if they are using a gym style environment that needs special logic.

        Args:
            env_name (str): Name of the environment
        """
        self._env = gym.make(env_name)

    def create_env_spec(self, env_name, **kwargs):
        """Function used to create the specification. Subclasses can override this method
        if they are using a gym style environment that needs special logic.

        Args:
            env_name (str): Name of the environment
        """
        if isinstance(self._env.observation_space, gym.spaces.Tuple):
            obs_spaces = self._env.observation_space.spaces
        else:
            obs_spaces = [self._env.observation_space]
        if isinstance(self._env.action_space, gym.spaces.Tuple):
            act_spaces = self._env.action_space.spaces
        else:
            act_spaces = [self._env.action_space]

        return EnvSpec(
            env_name=env_name,
            obs_dim=[space.shape for space in obs_spaces],
            act_dim=[space.n for space in act_spaces],
        )

    def reset(self):
        observation = self._env.reset()
        return observation, self._turn

    def step(self, action):
        observation, reward, done, info = self._env.step(action)
        self._turn = (self._turn + 1) % self._num_players
        return observation, reward, done, self._turn, info

    def render(self, mode="rgb_array"):
        return self._env.render(mode=mode)

    def seed(self, seed=None):
        self._env.seed(seed=seed)

    def close(self):
        self._env.close()



from hive.envs.base import BaseEnv, ParallelEnv
from hive.envs.env_spec import EnvSpec
from hive.envs.gym_env import GymEnv

try:
    from hive.envs.minigrid import MiniGridEnv
except ImportError:
    MiniGridEnv = None

try:
    from hive.envs.atari import AtariEnv
except ImportError:
    AtariEnv = None

try:
    from hive.envs.marlgrid import MarlGridEnv
except ImportError:
    MarlGridEnv = None

try:
    from hive.envs.minatar import MinAtarEnv
except ImportError:
    MinAtarEnv = None

try:
    from hive.envs.pettingzoo import PettingZooEnv
except ImportError:
    PettingZooEnv = None

from hive.utils.registry import registry

registry.register_all(
    BaseEnv,
    {
        "GymEnv": GymEnv,
        "MiniGridEnv": MiniGridEnv,
        "MarlGridEnv": MarlGridEnv,
        "AtariEnv": AtariEnv,
        "MinAtarEnv": MinAtarEnv,
        "PettingZooEnv": PettingZooEnv,
    },
)

get_env = getattr(registry, f"get_{BaseEnv.type_name()}")



import argparse
import copy

from hive import agents as agent_lib
from hive import envs
from hive.runners.base import Runner
from hive.runners.utils import TransitionInfo, load_config
from hive.utils import experiment, loggers, schedule, utils
from hive.utils.registry import get_parsed_args


class MultiAgentRunner(Runner):
    """Runner class used to implement a multiagent training loop."""

    def __init__(
        self,
        environment,
        agents,
        logger,
        experiment_manager,
        train_steps,
        test_frequency,
        test_episodes,
        stack_size,
        self_play,
        max_steps_per_episode=27000,
    ):
        """Initializes the Runner object.

        Args:
            environment (BaseEnv): Environment used in the training loop.
            agents (list[Agent]): List of agents that interact with the environment
            logger (ScheduledLogger): Logger object used to log metrics.
            experiment_manager (Experiment): Experiment object that saves the state of
                the training.
            train_steps (int): How many steps to train for. If this is -1, there is no
                limit for the number of training steps.
            test_frequency (int): After how many training steps to run testing
                episodes. If this is -1, testing is not run.
            test_episodes (int): How many episodes to run testing for.
            stack_size (int): The number of frames in an observation sent to an agent.
            max_steps_per_episode (int): The maximum number of steps to run an episode
                for.
        """
        super().__init__(
            environment,
            agents,
            logger,
            experiment_manager,
            train_steps,
            test_frequency,
            test_episodes,
            max_steps_per_episode,
        )
        self._transition_info = TransitionInfo(self._agents, stack_size)
        self._self_play = self_play

    def run_one_step(self, observation, turn, episode_metrics):
        """Run one step of the training loop.

        If it is the agent's first turn during the episode, do not run an update step.
        Otherwise, run an update step based on the previous action and accumulated
        reward since then.

        Args:
            observation: Current observation that the agent should create an action
                for.
            turn (int): Agent whose turn it is.
            episode_metrics (Metrics): Keeps track of metrics for current episode.
        """
        super().run_one_step(observation, turn, episode_metrics)
        agent = self._agents[turn]
        if self._transition_info.is_started(agent):
            info = self._transition_info.get_info(agent)
            if self._training:
                agent.update(copy.deepcopy(info))

            episode_metrics[agent.id]["reward"] += info["reward"]
            episode_metrics[agent.id]["episode_length"] += 1
            episode_metrics["full_episode_length"] += 1
        else:
            self._transition_info.start_agent(agent)

        stacked_observation = self._transition_info.get_stacked_state(
            agent, observation
        )
        action = agent.act(stacked_observation)
        next_observation, reward, done, turn, other_info = self._environment.step(
            action
        )
        self._transition_info.record_info(
            agent,
            {
                "observation": observation,
                "action": action,
                "info": other_info,
            },
        )
        if self._self_play:
            self._transition_info.record_info(
                agent,
                {
                    "agent_id": agent.id,
                },
            )
        self._transition_info.update_all_rewards(reward)
        return done, next_observation, turn

    def run_end_step(self, episode_metrics, done=True):
        """Run the final step of an episode.

        After an episode ends, iterate through agents and update then with the final
        step in the episode.

        Args:
            episode_metrics (Metrics): Keeps track of metrics for current episode.
            done (bool): Whether this step was terminal.

        """
        for agent in self._agents:
            if self._transition_info.is_started(agent):
                info = self._transition_info.get_info(agent, done=done)

                if self._training:
                    agent.update(info)
                episode_metrics[agent.id]["episode_length"] += 1
                episode_metrics["full_episode_length"] += 1
            episode_metrics[agent.id]["reward"] += info["reward"]

    def run_episode(self):
        """Run a single episode of the environment."""
        episode_metrics = self.create_episode_metrics()
        done = False
        observation, turn = self._environment.reset()
        self._transition_info.reset()
        steps = 0
        # Run the loop until the episode ends or times out
        while not done and steps < self._max_steps_per_episode:
            done, observation, turn = self.run_one_step(
                observation, turn, episode_metrics
            )
            steps += 1

        # Run the final update.
        self.run_end_step(episode_metrics, done)
        return episode_metrics


def set_up_experiment(config):
    """Returns a :py:class:`MultiAgentRunner` object based on the config and any
    command line arguments.

    Args:
        config: Configuration for experiment.
    """

    # Parses arguments from the command line.
    args = get_parsed_args(
        {
            "seed": int,
            "train_steps": int,
            "test_frequency": int,
            "test_episodes": int,
            "max_steps_per_episode": int,
            "stack_size": int,
            "resume": bool,
            "run_name": str,
            "save_dir": str,
            "self_play": bool,
            "num_agents": int,
        }
    )
    config.update(args)
    full_config = utils.Chomp(copy.deepcopy(config))

    if "seed" in config:
        utils.seeder.set_global_seed(config["seed"])

    # Set up environment
    environment, full_config["environment"] = envs.get_env(
        config["environment"], "environment"
    )
    env_spec = environment.env_spec

    # Set up loggers
    logger_config = config.get("loggers", {"name": "NullLogger"})
    if logger_config is None or len(logger_config) == 0:
        logger_config = {"name": "NullLogger"}
    if isinstance(logger_config, list):
        logger_config = {
            "name": "CompositeLogger",
            "kwargs": {"logger_list": logger_config},
        }

    logger, full_config["loggers"] = loggers.get_logger(logger_config, "loggers")

    # Set up agents
    agents = []
    full_config["agents"] = []
    num_agents = config["num_agents"] if config["self_play"] else len(config["agents"])
    for idx in range(num_agents):

        if not config["self_play"] or idx == 0:
            agent_config = config["agents"][idx]
            if config.get("stack_size", 1) > 1:
                agent_config["kwargs"]["obs_dim"] = (
                    config["stack_size"] * env_spec.obs_dim[idx][0],
                    *env_spec.obs_dim[idx][1:],
                )
            else:
                agent_config["kwargs"]["obs_dim"] = env_spec.obs_dim[idx]
            agent_config["kwargs"]["act_dim"] = env_spec.act_dim[idx]
            agent_config["kwargs"]["logger"] = logger

            if "replay_buffer" in agent_config["kwargs"]:
                replay_args = agent_config["kwargs"]["replay_buffer"]["kwargs"]
                replay_args["observation_shape"] = env_spec.obs_dim[idx]
            agent, full_agent_config = agent_lib.get_agent(
                agent_config, f"agents.{idx}"
            )
            agents.append(agent)
            full_config["agents"].append(full_agent_config)
        else:
            agents.append(copy.copy(agents[0]))
            agents[-1]._id = idx

    # Set up experiment manager
    saving_schedule, full_config["saving_schedule"] = schedule.get_schedule(
        config["saving_schedule"], "saving_schedule"
    )
    experiment_manager = experiment.Experiment(
        config["run_name"], config["save_dir"], saving_schedule
    )
    experiment_manager.register_experiment(
        config=full_config,
        logger=logger,
        agents=agents,
    )

    # Set up runner
    runner = MultiAgentRunner(
        environment,
        agents,
        logger,
        experiment_manager,
        config.get("train_steps", -1),
        config.get("test_frequency", -1),
        config.get("test_episodes", 1),
        config.get("stack_size", 1),
        config.get("self_play", False),
        config.get("max_steps_per_episode", 1e9),
    )
    if config.get("resume", False):
        runner.resume()

    return runner


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config")
    parser.add_argument("-p", "--preset-config")
    parser.add_argument("-a", "--agent-config")
    parser.add_argument("-e", "--env-config")
    parser.add_argument("-l", "--logger-config")
    args, _ = parser.parse_known_args()
    if args.config is None and args.preset_config is None:
        raise ValueError("Config needs to be provided")
    config = load_config(
        args.config,
        args.preset_config,
        args.agent_config,
        args.env_config,
        args.logger_config,
    )
    runner = set_up_experiment(config)
    runner.run_training()


if __name__ == "__main__":
    main()



import math

import torch
import torch.nn.functional as F
from torch import nn


class NoisyLinear(nn.Module):
    """NoisyLinear Layer. Implements the layer described in
    https://arxiv.org/abs/1706.10295."""

    def __init__(self, in_dim: int, out_dim: int, std_init: float = 0.5):
        """
        Args:
            in_dim (int): The dimension of the input.
            out_dim (int): The desired dimension of the output.
            std_init (float): The range for the initialization of the standard deviation of the
                weights.
        """
        super().__init__()
        self.in_features = in_dim
        self.out_features = out_dim
        self.std_init = std_init
        self.weight_mu = nn.Parameter(torch.empty(out_dim, in_dim))
        self.weight_sigma = nn.Parameter(torch.empty(out_dim, in_dim))
        self.register_buffer("weight_epsilon", torch.empty(out_dim, in_dim))
        self.bias_mu = nn.Parameter(torch.empty(out_dim))
        self.bias_sigma = nn.Parameter(torch.empty(out_dim))
        self.register_buffer("bias_epsilon", torch.empty(out_dim))
        self._reset_parameters()
        self._sample_noise()

    def _reset_parameters(self):
        mu_range = 1.0 / math.sqrt(self.in_features)
        self.weight_mu.data.uniform_(-mu_range, mu_range)
        self.weight_sigma.data.fill_(self.std_init / math.sqrt(self.in_features))
        self.bias_mu.data.uniform_(-mu_range, mu_range)
        self.bias_sigma.data.fill_(self.std_init / math.sqrt(self.out_features))

    def _scale_noise(self, size):
        x = torch.randn(size)
        return x.sign() * (x.abs().sqrt())

    def _sample_noise(self):
        epsilon_in = self._scale_noise(self.in_features)
        epsilon_out = self._scale_noise(self.out_features)
        weight_eps = epsilon_out.ger(epsilon_in)
        bias_eps = epsilon_out
        return weight_eps, bias_eps

    def forward(self, inp):
        if self.training:
            weight_eps, bias_eps = self._sample_noise()
            return F.linear(
                inp,
                self.weight_mu
                + self.weight_sigma * weight_eps.to(device=self.weight_sigma.device),
                self.bias_mu
                + self.bias_sigma * bias_eps.to(device=self.bias_sigma.device),
            )
        else:
            return F.linear(inp, self.weight_mu, self.bias_mu)



import argparse
import json
import logging
import os
import pickle
from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import cm

logging.basicConfig()


def find_single_run_data(run_folder):
    """Looks for a chomp logger data file in `run_folder` and it's subdirectories.
    Once it finds one, it loads the file and returns the data.

    Args:
        run_folder (str): Which folder to search for the chomp logger data.

    Returns:
        The Chomp object containing the logger data.
    """
    run_data_file = None
    for path, _, filenames in os.walk(run_folder):
        if "log_data.p" in filenames:
            run_data_file = os.path.join(path, "log_data.p")
            logging.info(f"Found run data at {run_data_file}")
            break
    if run_data_file is None:
        logging.info(f"Run data not found for {run_folder}")
        return

    with open(run_data_file, "rb") as f:
        run_data = pickle.load(f)

    return run_data


def find_all_runs_data(runs_folder):
    """Iterates through each directory in runs_folder, finds one chomp logger
    data file in each directory, and concatenates the data together under each
    key present in the data.

    Args:
        runs_folder (str): Folder which contains subfolders, each of from which one
            chomp logger data file is loaded.

    Returns:
        Dictionary with each key corresponding to a key in the chomp logger data files.
        The values are the folder
    """
    all_runs_data = defaultdict(lambda: [])
    for run_folder in os.listdir(runs_folder):
        full_run_folder = os.path.join(runs_folder, run_folder)
        if os.path.isdir(full_run_folder):
            run_data = find_single_run_data(full_run_folder)
            for key in run_data:
                all_runs_data[key].append(run_data[key])
    return all_runs_data


def find_all_experiments_data(experiments_folder, runs_folders):
    """Finds and loads all log data in a folder.

    Assuming the directory structure is as follows:

    ::

        experiment
        ├── config_1_runs
        |   ├──seed0/
        |   ├──seed1/
        |       .
        |       .
        |   └──seedn/
        ├── config_2_runs
        |   ├──seed0/
        |   ├──seed1/
        |       .
        |       .
        |   └──seedn/
        ├── config_3_runs
        |   ├──seed0/
        |   ├──seed1/
        |       .
        |       .
        |   └──seedn/

    Where there is some chomp logger data file under each seed directory, then
    passing "experiment" and the list of config folders ("config_1_runs",
    "config_2_runs", "config_3_runs"), will load all the data.

    Args:
        experiments_folder (str): Root folder with all experiments data.
        runs_folders (list[str]): List of folders under root folder to load data
            from.
    """
    data = {}
    for runs_folder in runs_folders:
        full_runs_folderpath = os.path.join(experiments_folder, runs_folder)
        data[runs_folder] = find_all_runs_data(full_runs_folderpath)
    return data


def standardize_data(
    experiment_data,
    x_key,
    y_key,
    num_sampled_points=1000,
    drop_last=True,
):
    """Extracts given keys from data, and standardizes the data across runs by sampling
    equally spaced points along x data, and interpolating y data of each run.

    Args:
        experiment_data: Data object in the format of
            :py:meth:`find_all_experiments_data`.
        x_key (str): Key for x axis.
        y_key (str): Key for y axis.
        num_sampled_points (int): How many points to sample along x axis.
        drop_last (bool): Whether to drop the last point in the data.
    """
    if drop_last:
        y_data = [data[0][:-1] for data in experiment_data[y_key]]
        x_data = [
            [x_datas[x_key] for x_datas in data[1][:-1]]
            for data in experiment_data[y_key]
        ]
    else:
        y_data = [data[0] for data in experiment_data[y_key]]
        x_data = [
            [x_datas[x_key] for x_datas in data[1]] for data in experiment_data[y_key]
        ]
    min_x = min([min(xs) for xs in x_data])
    max_x = max([max(xs) for xs in x_data])
    full_xs = np.linspace(min_x, max_x, num_sampled_points)
    interpolated_ys = np.array(
        [np.interp(full_xs, xs, ys) for (xs, ys) in zip(x_data, y_data)]
    )

    return full_xs, interpolated_ys


def find_and_standardize_data(
    experiments_folder, runs_folders, x_key, y_key, num_sampled_points, drop_last
):
    """Finds and standardizes the data in `experiments_folder`.

    Args:
        experiments_folder (str): Root folder with all experiments data.
        runs_folders (list[str]): List of folders under root folder to load data
            from.
        x_key (str): Key for x axis.
        y_key (str): Key for y axis.
        num_sampled_points (int): How many points to sample along x axis.
        drop_last (bool): Whether to drop the last point in the data.
    """
    if runs_folders is None:
        runs_folders = os.listdir(experiments_folder)
    data = find_all_experiments_data(experiments_folder, runs_folders)

    aggregated_xs, aggregated_ys = list(
        zip(
            *[
                standardize_data(
                    data[run_folder],
                    x_key,
                    y_key,
                    num_sampled_points=num_sampled_points,
                    drop_last=drop_last,
                )
                for run_folder in runs_folders
            ]
        )
    )

    return runs_folders, aggregated_xs, aggregated_ys


def generate_lineplot(
    x_datas,
    y_datas,
    smoothing_fn=None,
    line_labels=None,
    xlabel=None,
    ylabel=None,
    cmap_name=None,
    output_file="output.png",
):
    """Aggregates data and generates lineplot."""

    plt.figure()
    if cmap_name is None:
        cmap_name = "tab10" if len(x_datas) <= 10 else "tab20"
    cmap = cm.get_cmap(cmap_name)
    if line_labels is None:
        line_labels = [None] * len(x_datas)
    for idx, (x_data, y_data, line_label) in enumerate(
        zip(x_datas, y_datas, line_labels)
    ):
        mean_ys = np.mean(y_data, axis=0)
        std_ys = np.std(y_data, axis=0)
        if smoothing_fn is not None:
            mean_ys = smoothing_fn(mean_ys)
            std_ys = smoothing_fn(std_ys)
        plt.plot(x_data, mean_ys, label=line_label, color=cmap(idx))
        plt.fill_between(
            x_data, mean_ys - std_ys, mean_ys + std_ys, color=cmap(idx), alpha=0.1
        )
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.legend()
    plt.savefig(output_file)
    plt.close()


def plot_results(
    experiments_folder,
    x_key,
    y_key,
    runs_folders=None,
    drop_last=True,
    run_names=None,
    x_label=None,
    y_label=None,
    cmap_name=None,
    smoothing_fn=None,
    num_sampled_points=100,
    output_file="output.png",
):
    """Plots results."""
    runs_folders, aggregated_xs, aggregated_ys = find_and_standardize_data(
        experiments_folder,
        runs_folders,
        x_key,
        y_key,
        num_sampled_points,
        drop_last,
    )
    if run_names is None:
        run_names = runs_folders
    generate_lineplot(
        aggregated_xs,
        aggregated_ys,
        smoothing_fn=smoothing_fn,
        line_labels=run_names,
        xlabel=x_label,
        ylabel=y_label,
        cmap_name=cmap_name,
        output_file=output_file,
    )


def create_exponential_smoothing_fn(smoothing=0.1):
    def fn(values):
        values = np.array(values)
        return np.array(pd.DataFrame(values).ewm(alpha=1 - smoothing).mean()[0])

    return fn


def create_moving_average_smoothing_fn(running_average=10):
    def fn(values):
        return np.convolve(values, np.ones(running_average), "valid") / running_average

    return fn


def get_smoothing_fn(smoothing_fn, smoothing_fn_kwargs):
    if smoothing_fn == "exponential":
        return create_exponential_smoothing_fn(**smoothing_fn_kwargs)
    elif smoothing_fn == "moving_average":
        return create_moving_average_smoothing_fn(**smoothing_fn_kwargs)
    else:
        return None


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--experiments_folder", required=True)
    parser.add_argument("--x_key", required=True)
    parser.add_argument("--y_key", required=True)
    parser.add_argument("--runs_folders", nargs="+")
    parser.add_argument("--run_names", nargs="+")
    parser.add_argument("--x_label")
    parser.add_argument("--y_label")
    parser.add_argument("--cmap_name")
    parser.add_argument("--smoothing_fn", choices=["exponential", "moving_average"])
    parser.add_argument("--smoothing_fn_kwargs")
    parser.add_argument("--num_sampled_points", type=int, default=100)
    parser.add_argument("--num_sampled_points", type=int, default=100)
    parser.add_argument("--drop_last", action="store_true")

    args = parser.parse_args()
    if args.smoothing_fn is not None:
        if args.smoothing_fn_kwargs is not None:
            smoothing_fn_kwargs = json.loads(args.smoothing_fn_kwargs)
        smoothing_fn = get_smoothing_fn(args.smoothing_fn, smoothing_fn_kwargs)
    else:
        smoothing_fn = None
    plot_results(
        experiments_folder=args.experiments_folder,
        x_key=args.x_key,
        y_key=args.y_key,
        runs_folders=args.runs_folders,
        drop_last=args.drop_last,
        run_names=args.run_names,
        x_label=args.x_label,
        y_label=args.y_label,
        cmap_name=args.cmap_name,
        smoothing_fn=smoothing_fn,
        num_sampled_points=args.num_sampled_points,
        output_file=args.output_file,
    )



from importlib import import_module

import numpy as np

from hive.envs.base import BaseEnv
from hive.envs.env_spec import EnvSpec


class MinAtarEnv(BaseEnv):
    """
    Class for loading MinAtar environments. See https://github.com/kenjyoung/MinAtar.
    """

    def __init__(
        self,
        env_name,
        sticky_action_prob=0.1,
        difficulty_ramping=True,
    ):
        """
        Args:
            env_name (str): Name of the environment
            sticky_actions (bool): Whether to use sticky_actions as per
                Machado et al.
            difficulty_ramping (bool): Whether to periodically increase difficulty.
        """
        env_module = import_module("minatar.environments." + env_name)
        self.env_name = env_name
        self._env = env_module.Env(ramping=difficulty_ramping)
        self.n_channels = self._env.state_shape()[2]
        self.sticky_action_prob = sticky_action_prob
        self.last_action = 0
        self.visualized = False
        self.closed = False
        super().__init__(self.create_env_spec(env_name), num_players=1)

    def create_env_spec(self, env_name):
        obs_dim = tuple(self._env.state_shape())
        new_positions = [2, 0, 1]
        obs_dim = tuple(obs_dim[i] for i in new_positions)
        return EnvSpec(
            env_name=env_name,
            obs_dim=[obs_dim],
            act_dim=[6],
        )

    def reset(self):
        self._env.reset()
        return np.transpose(self._env.state(), [2, 1, 0]), 0

    def seed(self, seed=None):
        self._env.seed(seed=seed)

    def step(self, action=None):
        """
        Remarks:
            * Execute self.frame_skips steps taking the action in the the environment.
            * This may execute fewer than self.frame_skip steps in the environment,
              if the done state is reached.
            * Furthermore, in this case the returned observation should be ignored.
        """
        assert action is not None
        reward, done = self._env.act(action)
        reward = float(reward)
        info = {}
        observation = np.transpose(self._env.state(), [2, 1, 0])

        return observation, reward, done, None, info



import torch
from torch import nn

from hive.agents.qnets.mlp import MLPNetwork
from hive.agents.qnets.utils import calculate_output_dim


class ConvNetwork(nn.Module):
    """
    Basic convolutional neural network architecture. Applies a number of
    convolutional layers (each followed by a ReLU activation), and then
    feeds the output into an :py:class:`hive.agents.qnets.mlp.MLPNetwork`.

    Note, if :obj:`channels` is :const:`None`, the network created for the
    convolution portion of the architecture is simply an
    :py:class:`torch.nn.Identity` module. If :obj:`mlp_layers` is
    :const:`None`, the mlp portion of the architecture is an
    :py:class:`torch.nn.Identity` module.
    """

    def __init__(
        self,
        in_dim,
        channels=None,
        mlp_layers=None,
        kernel_sizes=1,
        strides=1,
        paddings=0,
        normalization_factor=255,
        noisy=False,
        std_init=0.5,
    ):
        """
        Args:
            in_dim (tuple): The tuple of observations dimension (channels, width,
                height).
            channels (list): The size of output channel for each convolutional layer.
            mlp_layers (list): The number of neurons for each mlp layer after the
                convolutional layers.
            kernel_sizes (list | int): The kernel size for each convolutional layer
            strides (list | int): The stride used for each convolutional layer.
            paddings (list | int): The size of the padding used for each convolutional
                layer.
            normalization_factor (float | int): What the input is divided by before
                the forward pass of the network.
            noisy (bool): Whether the MLP part of the network will use
                :py:class:`~hive.agents.qnets.noisy_linear.NoisyLinear` layers or
                :py:class:`torch.nn.Linear` layers.
            std_init (float): The range for the initialization of the standard
                deviation of the weights in
                :py:class:`~hive.agents.qnets.noisy_linear.NoisyLinear`.
        """
        super().__init__()
        self._normalization_factor = normalization_factor
        if channels is not None:
            if isinstance(kernel_sizes, int):
                kernel_sizes = [kernel_sizes] * len(channels)
            if isinstance(strides, int):
                strides = [strides] * len(channels)
            if isinstance(paddings, int):
                paddings = [paddings] * len(channels)

            if not all(
                len(x) == len(channels) for x in [kernel_sizes, strides, paddings]
            ):
                raise ValueError("The lengths of the parameter lists must be the same")

            # Convolutional Layers
            channels.insert(0, in_dim[0])
            conv_seq = []
            for i in range(0, len(channels) - 1):
                conv_seq.append(
                    torch.nn.Conv2d(
                        in_channels=channels[i],
                        out_channels=channels[i + 1],
                        kernel_size=kernel_sizes[i],
                        stride=strides[i],
                        padding=paddings[i],
                    )
                )
                conv_seq.append(torch.nn.ReLU())
            self.conv = torch.nn.Sequential(*conv_seq)
        else:
            self.conv = torch.nn.Identity()

        if mlp_layers is not None:
            # MLP Layers
            conv_output_size = calculate_output_dim(self.conv, in_dim)
            self.mlp = MLPNetwork(
                conv_output_size, mlp_layers, noisy=noisy, std_init=std_init
            )
        else:
            self.mlp = torch.nn.Identity()

    def forward(self, x):
        if len(x.shape) == 3:
            x = x.unsqueeze(0)
        elif len(x.shape) == 5:
            x = x.reshape(x.size(0), -1, x.size(-2), x.size(-1))
        x = x.float()
        x = x / self._normalization_factor
        x = self.conv(x)
        x = self.mlp(x)
        return x



from gym.envs.registration import register
from marlgrid.envs import register_marl_env

from hive.envs.marlgrid.ma_envs.checkers import CheckersMultiGrid
from hive.envs.marlgrid.ma_envs.pursuit import PursuitMultiGrid
from hive.envs.marlgrid.ma_envs.switch import SwitchMultiGrid

register_marl_env(
    "MarlGrid-2AgentCheckers8x8-v0",
    CheckersMultiGrid,
    n_agents=2,
    grid_size=8,
    view_size=5,  # Needs to be same as grid_size if full_obs = True.
    env_kwargs={"max_steps": 100, "full_obs": False},
)

register_marl_env(
    "MarlGrid-2Agent1RandomPursuit8x8-v0",
    PursuitMultiGrid,
    n_agents=3,
    grid_size=8,
    view_size=5,  # Needs to be same as grid_size if full_obs = True.
    env_kwargs={"max_steps": 500, "full_obs": False},
)

register_marl_env(
    "MarlGrid-2AgentSwitch8x8-v0",
    SwitchMultiGrid,
    n_agents=2,
    grid_size=12,
    view_size=12,  # Needs to be same as grid_size if full_obs = True.
    env_kwargs={"max_steps": 500, "full_obs": True},
)



import numpy as np
from marlgrid.base import MultiGrid
from marlgrid.objects import Goal, GridAgent, Lava, Wall

from hive.envs.marlgrid.ma_envs.base import MultiGridEnvHive


class PursuitMultiGrid(MultiGridEnvHive):
    """
    Pursuit–Evasion environment based on Gupta et al. 2017

    "The pursuit-evasion domain consists of two sets of agents: evaders and pursuers.
    The evaders are trying to avoid pursuers, while the pursuers are
    trying to catch the evaders. The pursuers receive a reward of 5.0 when
    they surround an evader or corner the agent"
    """

    metadata = {}

    def _gen_grid(self, width, height):
        self.grid = MultiGrid((width, height))
        self.grid.wall_rect(0, 0, width, height)
        self.ghost_mode = False

    def reset(self, **kwargs):
        obs = super().reset()
        for ag_idx, _ in enumerate(obs):
            obs[ag_idx] = np.array(obs[ag_idx], dtype=np.uint8)
        return obs

    def step(self, actions):
        # Spawn agents if it's time.
        for agent in self.agents:
            if (
                not agent.active
                and not agent.done
                and self.step_count >= agent.spawn_delay
            ):
                self.place_obj(agent, **self.agent_spawn_kwargs)
                agent.activate()

        num_learning_agents = len(actions)
        num_rand_agents = self.num_agents - len(actions)

        step_rewards = np.zeros((num_learning_agents), dtype=np.float)

        self.step_count += 1
        for i in range(num_learning_agents, self.num_agents):
            actions.append(self.action_space[i].sample())
        iter_agents = list(enumerate(zip(self.agents, actions)))
        iter_order = np.arange(len(iter_agents))
        for shuffled_ix in iter_order:
            agent_no, (agent, action) = iter_agents[shuffled_ix]
            agent.step_reward = 0

            if agent.active:

                cur_pos = agent.pos[:]
                cur_cell = self.grid.get(*cur_pos)
                fwd_pos = agent.front_pos[:]
                fwd_cell = self.grid.get(*fwd_pos)
                agent_moved = False
                bot_pos = agent.pos + np.array([0, -1])
                bot_cell = self.grid.get(*bot_pos)
                abov_pos = agent.pos + np.array([0, +1])
                abov_cell = self.grid.get(*abov_pos)
                left_pos = agent.pos + np.array([-1, 0])
                left_cell = self.grid.get(*left_pos)
                right_pos = agent.pos + np.array([+1, 0])
                right_cell = self.grid.get(*right_pos)

                w = 0
                a = 0
                surrounding_cells = [bot_cell, abov_cell, left_cell, right_cell]
                if agent_no == len(self.agents) - num_rand_agents:
                    for cell in surrounding_cells:
                        if isinstance(cell, GridAgent):
                            a += 1
                        if isinstance(cell, Wall):
                            w += 1

                    if a == len(self.agents) - num_rand_agents or (w == 2 and a == 2):
                        step_rewards[:] += np.array(
                            [5] * (len(self.agents) - num_rand_agents)
                        )
                        for agent in self.agents:
                            agent.done = True

                # Rotate left
                if action == agent.actions.left:
                    agent.dir = (agent.dir - 1) % 4

                # Rotate right
                elif action == agent.actions.right:
                    agent.dir = (agent.dir + 1) % 4

                # Move forward
                elif action == agent.actions.forward:
                    # Under the follow conditions, the agent can move forward.
                    can_move = fwd_cell is None or fwd_cell.can_overlap()
                    if self.ghost_mode is False and isinstance(fwd_cell, GridAgent):
                        can_move = False

                    if can_move:
                        agent_moved = True
                        # Add agent to new cell
                        if fwd_cell is None:
                            self.grid.set(*fwd_pos, agent)
                            agent.pos = fwd_pos
                        else:
                            fwd_cell.agents.append(agent)
                            agent.pos = fwd_pos

                        # Remove agent from old cell
                        if cur_cell == agent:
                            self.grid.set(*cur_pos, None)
                        else:
                            assert cur_cell.can_overlap()
                            cur_cell.agents.remove(agent)

                        # Add agent's agents to old cell
                        for left_behind in agent.agents:
                            cur_obj = self.grid.get(*cur_pos)
                            if cur_obj is None:
                                self.grid.set(*cur_pos, left_behind)
                            elif cur_obj.can_overlap():
                                cur_obj.agents.append(left_behind)
                            else:
                                raise ValueError(
                                    "How was agent there in teh first place?"
                                )

                        # After moving, the agent shouldn't contain any other agents.
                        agent.agents = []

                        # Rewards can be got iff. fwd_cell has a "get_reward" method
                        if hasattr(fwd_cell, "get_reward"):
                            rwd = fwd_cell.get_reward(agent)
                            if bool(self.reward_decay):
                                rwd *= 1.0 - 0.9 * (self.step_count / self.max_steps)
                            step_rewards[agent_no] += rwd
                            agent.reward(rwd)

                        if isinstance(fwd_cell, (Lava, Goal)):
                            agent.done = True

                # Pick up an object
                elif action == agent.actions.pickup:
                    if fwd_cell and fwd_cell.can_pickup():
                        if agent.carrying is None:
                            agent.carrying = fwd_cell
                            agent.carrying.cur_pos = np.array([-1, -1])
                            self.grid.set(*fwd_pos, None)
                    else:
                        pass

                # Drop an object
                elif action == agent.actions.drop:
                    if not fwd_cell and agent.carrying:
                        self.grid.set(*fwd_pos, agent.carrying)
                        agent.carrying.cur_pos = fwd_pos
                        agent.carrying = None
                    else:
                        pass

                # Toggle/activate an object
                elif action == agent.actions.toggle:
                    if fwd_cell:
                        wasted = bool(fwd_cell.toggle(agent, fwd_pos))
                    else:
                        pass

                # Done action (not used by default)
                elif action == agent.actions.done:
                    pass

                else:
                    raise ValueError(f"Environment can't handle action {action}.")

                agent.on_step(fwd_cell if agent_moved else None)

        # If any of the agents individually are "done" (hit lava or in some cases a goal)
        # but the env requires respawning, then respawn those agents.
        for agent in self.agents:
            if agent.done:
                if self.respawn:
                    resting_place_obj = self.grid.get(*agent.pos)
                    if resting_place_obj == agent:
                        if agent.agents:
                            self.grid.set(*agent.pos, agent.agents[0])
                            agent.agents[0].agents += agent.agents[1:]
                        else:
                            self.grid.set(*agent.pos, None)
                    else:
                        resting_place_obj.agents.remove(agent)
                        resting_place_obj.agents += agent.agents[:]
                        agent.agents = []

                    agent.reset(new_episode=False)
                    self.place_obj(agent, **self.agent_spawn_kwargs)
                    agent.activate()
                else:  # if the agent shouldn't be respawned, then deactivate it.
                    agent.deactivate()

        # The episode overall is done if all the agents are done,
        # or if it exceeds the step limit.
        done = (self.step_count >= self.max_steps) or all(
            [agent.done for agent in self.agents[:num_learning_agents]]
        )

        obs = [
            np.asarray(self.gen_agent_obs(agent), dtype=np.uint8)
            for agent in self.agents[:num_learning_agents]
        ]

        return obs, step_rewards, done, {}



import math

import torch

from hive.utils.registry import registry
from hive.utils.utils import CallableType


def calculate_output_dim(net, input_shape):
    """Calculates the resulting output shape for a given input shape and network.

    Args:
        net (torch.nn.Module): The network which you want to calculate the output
            dimension for.
        input_shape (int | tuple[int]): The shape of the input being fed into the
            :obj:`net`. Batch dimension should not be included.
    Returns:
        The shape of the output of a network given an input shape.
        Batch dimension is not included.
    """
    if isinstance(input_shape, int):
        input_shape = (input_shape,)
    placeholder = torch.zeros((0,) + tuple(input_shape))
    output = net(placeholder)
    return output.size()[1:]


def create_init_weights_fn(initialization_fn):
    """Returns a function that wraps :func:`initialization_function` and applies
    it to modules that have the :attr:`weight` attribute.

    Args:
        initialization_fn (callable): A function that takes in a tensor and
            initializes it.
    Returns:
        Function that takes in PyTorch modules and initializes their weights.
        Can be used as follows:

        .. code-block:: python

            init_fn = create_init_weights_fn(variance_scaling_)
            network.apply(init_fn)
    """
    if initialization_fn is not None:

        def init_weights(m):
            if hasattr(m, "weight"):
                initialization_fn(m.weight)

        return init_weights
    else:
        return lambda m: None


def calculate_correct_fan(tensor, mode):
    """Calculate fan of tensor.

    Args:
        tensor (torch.Tensor): Tensor to calculate fan of.
        mode (str): Which type of fan to compute. Must be one of `"fan_in"`,
            `"fan_out"`, and `"fan_avg"`.
    Returns:
        Fan of the tensor based on the mode.
    """
    fan_in, fan_out = torch.nn.init._calculate_fan_in_and_fan_out(tensor)
    if mode == "fan_in":
        return fan_in
    elif mode == "fan_out":
        return fan_out
    elif mode == "fan_avg":
        return (fan_in + fan_out) / 2
    else:
        raise ValueError(f"Fan mode {mode} not supported")


def variance_scaling_(tensor, scale=1.0, mode="fan_in", distribution="uniform"):
    """Implements the :py:class:`tf.keras.initializers.VarianceScaling`
    initializer in PyTorch.

    Args:
        tensor (torch.Tensor): Tensor to initialize.
        scale (float): Scaling factor (must be positive).
        mode (str): Must be one of `"fan_in"`, `"fan_out"`, and `"fan_avg"`.
        distribution: Random distribution to use, must be one of
            "truncated_normal", "untruncated_normal" and "uniform".
    Returns:
        Initialized tensor.
    """
    fan = calculate_correct_fan(tensor, mode)
    scale /= fan
    if distribution == "truncated_normal":
        # constant from scipy.stats.truncnorm.std(a=-2, b=2, loc=0., scale=1.)
        stddev = math.sqrt(scale) / 0.87962566103423978
        return torch.nn.init.trunc_normal_(tensor, 0.0, stddev, -2 * stddev, 2 * stddev)
    elif distribution == "untruncated_normal":
        stddev = math.sqrt(scale)
        return torch.nn.init.normal_(tensor, 0.0, stddev)
    elif distribution == "uniform":
        limit = math.sqrt(3.0 * scale)
        return torch.nn.init.uniform_(tensor, -limit, limit)
    else:
        raise ValueError(f"Distribution {distribution} not supported")


class InitializationFn(CallableType):
    """A wrapper for callables that produce initialization functions.

    These wrapped callables can be partially initialized through configuration
    files or command line arguments.
    """

    @classmethod
    def type_name(cls):
        """
        Returns:
            "init_fn"
        """
        return "init_fn"


registry.register_all(
    InitializationFn,
    {
        "uniform": InitializationFn(torch.nn.init.uniform_),
        "normal": InitializationFn(torch.nn.init.normal_),
        "constant": InitializationFn(torch.nn.init.constant_),
        "ones": InitializationFn(torch.nn.init.ones_),
        "zeros": InitializationFn(torch.nn.init.zeros_),
        "eye": InitializationFn(torch.nn.init.eye_),
        "dirac": InitializationFn(torch.nn.init.dirac_),
        "xavier_uniform": InitializationFn(torch.nn.init.xavier_uniform_),
        "xavier_normal": InitializationFn(torch.nn.init.xavier_normal_),
        "kaiming_uniform": InitializationFn(torch.nn.init.kaiming_uniform_),
        "kaiming_normal": InitializationFn(torch.nn.init.kaiming_normal_),
        "orthogonal": InitializationFn(torch.nn.init.orthogonal_),
        "sparse": InitializationFn(torch.nn.init.sparse_),
        "variance_scaling": InitializationFn(variance_scaling_),
    },
)

get_optimizer_fn = getattr(registry, f"get_{InitializationFn.type_name()}")



import gym_minigrid
from gym_minigrid.wrappers import (
    FlatObsWrapper,
    FullyObsWrapper,
    ImgObsWrapper,
    RGBImgObsWrapper,
    RGBImgPartialObsWrapper,
)

from hive.envs.gym_env import GymEnv
from hive.envs.wrappers.gym_wrappers import FlattenWrapper, PermuteImageWrapper


class MiniGridEnv(GymEnv):
    """
    Class for loading MiniGrid environments (https://github.com/maximecb/gym-minigrid).
    """

    def create_env(
        self,
        env_name,
        rgb_obs=True,
        flattened_obs=False,
        fully_observable=True,
        use_mission=False,
    ):
        """
        Args:
            env_name (str): Name of the environment.
            rgb_obs (bool): True if observations should be rgb-like images
            flattened_obs (bool): True for flattening the observation into one
                dimensional vector.
            fully_observable (bool): Whether to make the environment fully observable.
            use_mission (bool): Whether mission should be in the observation, in which
                case if using non-flattened grid, the observation is a dict of keys,
                image and mission. If using flattened observations, then the
                observation has the mission encoded in it.
        """

        super().create_env(env_name)

        if fully_observable:
            self._env = FullyObsWrapper(self._env)
            if rgb_obs:
                self._env = RGBImgObsWrapper(self._env)
        elif rgb_obs:
            self._env = RGBImgPartialObsWrapper(self._env)

        if not use_mission:
            self._env = PermuteImageWrapper(ImgObsWrapper(self._env))
            if flattened_obs:
                self._env = FlattenWrapper(self._env)
        elif flattened_obs:
            # Encode the mission into observation vector
            self._env = FlatObsWrapper(self._env)

    def render(self, mode="rgb_array"):
        # TODO
        pass



from typing import Dict, Tuple

import numpy as np

from hive.replays.prioritized_replay import PrioritizedReplayBuffer


class LegalMovesBuffer(PrioritizedReplayBuffer):
    """A Prioritized Replay buffer for the games like Hanabi with legal moves
    which need to add next_action_mask to the batch.
    """

    def __init__(
        self,
        capacity: int,
        beta: float = 0.5,
        stack_size: int = 1,
        n_step: int = 1,
        gamma: float = 0.9,
        observation_shape: Tuple = (),
        observation_dtype: type = np.uint8,
        action_shape: Tuple = (),
        action_dtype: type = np.int8,
        reward_shape: Tuple = (),
        reward_dtype: type = np.float32,
        extra_storage_types: Dict = None,
        action_dim: int = None,
        num_players_sharing_buffer: int = None,
    ):
        if extra_storage_types is None:
            extra_storage_types = {}
        extra_storage_types["action_mask"] = (np.float, [action_dim])
        super().__init__(
            capacity=capacity,
            stack_size=stack_size,
            beta=beta,
            n_step=n_step,
            gamma=gamma,
            observation_shape=observation_shape,
            observation_dtype=observation_dtype,
            action_shape=action_shape,
            action_dtype=action_dtype,
            reward_shape=reward_shape,
            reward_dtype=reward_dtype,
            extra_storage_types=extra_storage_types,
            num_players_sharing_buffer=num_players_sharing_buffer,
        )

    def sample(self, batch_size):
        """Sample transitions from the buffer.
        Adding next_action_mask to the batch for environments with legal moves.
        """

        batch = super().sample(batch_size)

        batch["next_action_mask"] = self._get_from_storage(
            "action_mask",
            batch["indices"] + batch["trajectory_lengths"] - self._stack_size + 1,
            num_to_access=1,
        )

        return batch



import operator
from functools import reduce

import gym
import numpy as np


class FlattenWrapper(gym.core.ObservationWrapper):
    """
    Flatten the observation to one dimensional vector.
    """

    def __init__(self, env):
        super().__init__(env)

        if isinstance(env.observation_space, gym.spaces.Tuple):
            self.observation_space = gym.spaces.Tuple(
                tuple(
                    gym.spaces.Box(
                        low=space.low.flatten(),
                        high=space.high.flatten(),
                        shape=(reduce(operator.mul, space.shape, 1),),
                        dtype=space.dtype,
                    )
                    for space in env.observation_space
                )
            )
            self._is_tuple = True
        else:
            self.observation_space = gym.spaces.Box(
                low=env.observation_space.low.flatten(),
                high=env.observation_space.high.flatten(),
                shape=(reduce(operator.mul, env.observation_space.shape, 1),),
                dtype=env.observation_space.dtype,
            )
            self._is_tuple = False

    def observation(self, obs):
        if self._is_tuple:
            return tuple(o.flatten() for o in obs)
        else:
            return obs.flatten()


class PermuteImageWrapper(gym.core.ObservationWrapper):
    """Changes the image format from HWC to CHW"""

    def __init__(self, env):
        super().__init__(env)

        if isinstance(env.observation_space, gym.spaces.Tuple):
            self.observation_space = gym.spaces.Tuple(
                tuple(
                    gym.spaces.Box(
                        low=np.transpose(space.low, [2, 1, 0]),
                        high=np.transpose(space.high, [2, 1, 0]),
                        shape=(space.shape[-1],) + space.shape[:-1],
                        dtype=space.dtype,
                    )
                    for space in env.observation_space
                )
            )
            self._is_tuple = True
        else:
            self.observation_space = gym.spaces.Box(
                low=np.transpose(env.observation_space.low, [2, 1, 0]),
                high=np.transpose(env.observation_space.high, [2, 1, 0]),
                shape=(env.observation_space.shape[-1],)
                + env.observation_space.shape[:-1],
                dtype=env.observation_space.dtype,
            )
            self._is_tuple = False

    def observation(self, obs):
        if self._is_tuple:
            return tuple(np.transpose(o, [2, 1, 0]) for o in obs)
        else:
            return np.transpose(obs, [2, 1, 0])



import os
import pickle
import random

import numpy as np
import torch

from hive.utils.registry import CallableType

PACKAGE_ROOT = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def create_folder(folder):
    """Creates a folder.

    Args:
        folder (str): Folder to create.
    """
    if not os.path.exists(folder):
        os.makedirs(folder)


class Seeder:
    """Class used to manage seeding in RLHive. It sets the seed for all the frameworks
    that RLHive currently uses. It also deterministically provides new seeds based on
    the global seed, in case any other objects in RLHive (such as the agents) need
    their own seed.
    """

    def __init__(self):
        self._seed = 0
        self._current_seed = 0

    def set_global_seed(self, seed):
        """This reduces some sources of randomness in experiments. To get reproducible
        results, you must run on the same machine and set the environment variable
        CUBLAS_WORKSPACE_CONFIG to ":4096:8" or ":16:8" before starting the experiment.

        Args:
            seed (int): Global seed.
        """
        self._seed = seed
        self._current_seed = seed
        torch.manual_seed(self._seed)
        random.seed(self._seed)
        np.random.seed(self._seed)
        torch.backends.cudnn.benchmark = False
        torch.use_deterministic_algorithms(True)

    def get_new_seed(self):
        """Each time it is called, it increments the current_seed and returns it."""
        self._current_seed += 1
        return self._current_seed


seeder = Seeder()


class Chomp(dict):
    """An extension of the dictionary class that allows for accessing through dot
    notation and easy saving/loading.
    """

    def __getattr__(self, k):
        if k not in self:
            raise AttributeError()
        return self.__getitem__(k)

    def __setattr__(self, k, v):
        self.__setitem__(k, v)

    def save(self, filename):
        """Saves the object using pickle.

        Args:
            filename (str): Filename to save object.
        """
        pickle.dump(self, open(filename, "wb"))

    def load(self, filename):
        """Loads the object.

        Args:
            filename (str): Where to load object from.
        """

        self.clear()
        self.update(pickle.load(open(filename, "rb")))


class OptimizerFn(CallableType):
    """A wrapper for callables that produce optimizer functions.

    These wrapped callables can be partially initialized through configuration
    files or command line arguments.
    """

    @classmethod
    def type_name(cls):
        """
        Returns:
            "optimizer_fn"
        """
        return "optimizer_fn"


class LossFn(CallableType):
    """A wrapper for callables that produce loss functions.

    These wrapped callables can be partially initialized through configuration
    files or command line arguments.
    """

    @classmethod
    def type_name(cls):
        """
        Returns:
            "loss_fn"
        """
        return "loss_fn"



import numpy as np
from marlgrid.base import MultiGrid
from marlgrid.objects import Goal, GridAgent

from hive.envs.marlgrid.ma_envs.base import MultiGridEnvHive


class CheckersMultiGrid(MultiGridEnvHive):
    """
    Checkers environment based on sunehag et al. 2017

    "... The map contains apples and lemons. The first player is very sensitive and scores 10 for
    the team for an apple (green square) and −10 for a lemon (orange square).
    The second, less sensitive player scores 1 for the team for an apple and −1 for a lemon.
    There is a wall of lemons between the players and the apples.
    Apples and lemons disappear when collected.
    The environment resets when all apples are eaten or maximum number of steps is reached.
    """

    def _gen_grid(self, width, height):
        self.num_rows = 3
        self.grid = MultiGrid((width, height))
        self.grid.wall_rect(0, 0, width, height)
        apple = Goal(color="green", reward=10)
        orange = Goal(color="red", reward=-10)
        self.num_remained_apples = 0
        for j in range(self.num_rows):
            oranges_loc = [2 * i + 1 + j % 2 for i in range(width // 2 - 1)]
            apples_loc = [2 * i + 1 + (j + 1) % 2 for i in range(width // 2 - 1)]
            for orange_loc in oranges_loc:
                self.put_obj(orange, orange_loc, j + 1)

            for apple_loc in apples_loc:
                self.put_obj(apple, apple_loc, j + 1)
                self.num_remained_apples += 1

        self.agent_spawn_kwargs = {}
        self.ghost_mode = False

    def reset(self, **kwargs):
        for agent in self.agents:
            agent.agents = []
            agent.reset(new_episode=True)

        self._gen_grid(self.width, self.height)

        for agent in self.agents:
            if agent.spawn_delay == 0:
                self.place_obj(
                    agent,
                    top=(0, self.num_rows + 1),
                    size=(self.width, self.height - self.num_rows - 1),
                    **self.agent_spawn_kwargs,
                )
                agent.activate()

        self.step_count = 0
        obs = self.gen_obs()
        for ag_idx, _ in enumerate(obs):
            obs[ag_idx] = np.array(obs[ag_idx], dtype=np.uint8)
        return obs

    def step(self, actions):
        # Spawn agents if it's time.
        for agent in self.agents:
            if (
                not agent.active
                and not agent.done
                and self.step_count >= agent.spawn_delay
            ):
                self.place_obj(agent, **self.agent_spawn_kwargs)
                agent.activate()

        assert len(actions) == len(self.agents)

        step_rewards = np.zeros((len(self.agents)), dtype=np.float)

        self.step_count += 1

        iter_agents = list(enumerate(zip(self.agents, actions)))
        iter_order = np.arange(len(iter_agents))
        self.np_random.shuffle(iter_order)
        for shuffled_ix in iter_order:
            agent_no, (agent, action) = iter_agents[shuffled_ix]
            agent.step_reward = 0

            if agent.active:

                cur_pos = agent.pos[:]
                cur_cell = self.grid.get(*cur_pos)
                fwd_pos = agent.front_pos[:]
                fwd_cell = self.grid.get(*fwd_pos)
                agent_moved = False

                # Rotate left
                if action == agent.actions.left:
                    agent.dir = (agent.dir - 1) % 4

                # Rotate right
                elif action == agent.actions.right:
                    agent.dir = (agent.dir + 1) % 4

                # Move forward
                elif action == agent.actions.forward:
                    # Under the follow conditions, the agent can move forward.
                    can_move = fwd_cell is None or fwd_cell.can_overlap()
                    if self.ghost_mode is False and isinstance(fwd_cell, GridAgent):
                        can_move = False

                    if can_move:
                        agent_moved = True
                        # Add agent to new cell
                        if fwd_cell is None or isinstance(fwd_cell, Goal):
                            self.grid.set(*fwd_pos, agent)
                            agent.pos = fwd_pos
                        else:
                            fwd_cell.agents.append(agent)
                            agent.pos = fwd_pos

                        # Remove agent from old cell
                        if cur_cell == agent:
                            self.grid.set(*cur_pos, None)
                        else:
                            assert cur_cell.can_overlap()
                            cur_cell.agents.remove(agent)

                        # Add agent's agents to old cell
                        for left_behind in agent.agents:
                            cur_obj = self.grid.get(*cur_pos)
                            if cur_obj is None:
                                self.grid.set(*cur_pos, left_behind)
                            elif cur_obj.can_overlap():
                                cur_obj.agents.append(left_behind)
                            else:
                                raise ValueError(
                                    "How was agent there in the first place?"
                                )

                        # After moving, the agent shouldn't contain any other agents.
                        agent.agents = []

                        # Rewards can be got iff. fwd_cell has a "get_reward" method
                        if hasattr(fwd_cell, "get_reward"):
                            rwd = fwd_cell.get_reward(agent)

                            # Modify the reward for less sensitive agent
                            if agent_no == 0:
                                rwd /= 10
                            if bool(self.reward_decay):
                                rwd *= 1.0 - 0.9 * (self.step_count / self.max_steps)
                            step_rewards[agent_no] += rwd
                            agent.reward(rwd)
                            if rwd > 0:
                                self.num_remained_apples -= 1

                # Pick up an object
                elif action == agent.actions.pickup:
                    if fwd_cell and fwd_cell.can_pickup() and agent.carrying is None:
                        agent.carrying = fwd_cell
                        agent.carrying.cur_pos = np.array([-1, -1])
                        self.grid.set(*fwd_pos, None)

                # Drop an object
                elif action == agent.actions.drop:
                    if not fwd_cell and agent.carrying:
                        self.grid.set(*fwd_pos, agent.carrying)
                        agent.carrying.cur_pos = fwd_pos
                        agent.carrying = None

                # Toggle/activate an object
                elif action == agent.actions.toggle:
                    if fwd_cell:
                        wasted = bool(fwd_cell.toggle(agent, fwd_pos))

                # Done action (not used by default)
                elif action == agent.actions.done:
                    pass

                else:
                    raise ValueError(f"Environment can't handle action {action}.")

                agent.on_step(fwd_cell if agent_moved else None)

        # If any of the agents individually are "done" (hit lava or in some cases a goal)
        # but the env requires respawning, then respawn those agents.
        for agent in self.agents:
            if agent.done:
                if self.respawn:
                    resting_place_obj = self.grid.get(*agent.pos)
                    if resting_place_obj == agent:
                        if agent.agents:
                            self.grid.set(*agent.pos, agent.agents[0])
                            agent.agents[0].agents += agent.agents[1:]
                        else:
                            self.grid.set(*agent.pos, None)
                    else:
                        resting_place_obj.agents.remove(agent)
                        resting_place_obj.agents += agent.agents[:]
                        agent.agents = []

                    agent.reset(new_episode=False)
                    self.place_obj(agent, **self.agent_spawn_kwargs)
                    agent.activate()
                else:  # if the agent shouldn't be respawned, then deactivate it.
                    agent.deactivate()

        # The episode overall is done if all the agents are done,
        # or if it exceeds the step limit or all the apples are collected.
        done = (
            (self.step_count >= self.max_steps)
            or all([agent.done for agent in self.agents])
            or self.num_remained_apples == 0
        )

        obs = [
            np.asarray(self.gen_agent_obs(agent), dtype=np.uint8)
            for agent in self.agents
        ]

        # Team reward
        step_rewards = np.array([np.sum(step_rewards) for _ in self.agents])

        return obs, step_rewards, done, {}



import gym
import numpy as np
from marlgrid.base import MultiGrid, MultiGridEnv, rotate_grid
from marlgrid.rendering import SimpleImageViewer

TILE_PIXELS = 32


class MultiGridEnvHive(MultiGridEnv):
    def __init__(
        self,
        agents,
        grid_size=None,
        width=None,
        height=None,
        max_steps=100,
        reward_decay=True,
        seed=1337,
        respawn=False,
        ghost_mode=True,
        full_obs=False,
        agent_spawn_kwargs={},
    ):
        self._full_obs = full_obs
        super().__init__(
            agents,
            grid_size,
            width,
            height,
            max_steps,
            reward_decay,
            seed,
            respawn,
            ghost_mode,
            agent_spawn_kwargs,
        )

    def gen_obs_grid(self, agent):
        # If the agent is inactive, return an empty grid and a visibility mask that hides everything.
        if not agent.active:
            # below, not sure orientation is correct but as of 6/27/2020 that doesn't matter because
            # agent views are usually square and this grid won't be used for anything.
            grid = MultiGrid(
                (agent.view_size, agent.view_size), orientation=agent.dir + 1
            )
            vis_mask = np.zeros((agent.view_size, agent.view_size), dtype=np.bool)
            return grid, vis_mask

        if self._full_obs:
            topX, topY, botX, botY = 0, 0, self.width, self.height
            grid = self.grid.slice(topX, topY, self.width, self.height, rot_k=0)
            vis_mask = np.ones((self.width, self.height), dtype=bool)
        else:
            topX, topY, botX, botY = agent.get_view_exts()
            grid = self.grid.slice(
                topX, topY, agent.view_size, agent.view_size, rot_k=agent.dir + 1
            )
            # Process occluders and visibility
            # Note that this incurs some slight performance cost
            vis_mask = agent.process_vis(grid.opacity)

        # Warning about the rest of the function:
        #  Allows masking away objects that the agent isn't supposed to see.
        #  But breaks consistency between the states of the grid objects in the parial views
        #   and the grid objects overall.
        if len(getattr(agent, "hide_item_types", [])) > 0:
            for i in range(grid.width):
                for j in range(grid.height):
                    item = grid.get(i, j)
                    if (
                        (item is not None)
                        and (item is not agent)
                        and (item.type in agent.hide_item_types)
                    ):
                        if len(item.agents) > 0:
                            grid.set(i, j, item.agents[0])
                        else:
                            grid.set(i, j, None)

        return grid, vis_mask

    def render(
        self,
        mode="human",
        close=False,
        highlight=True,
        tile_size=TILE_PIXELS,
        show_agent_views=True,
        max_agents_per_col=3,
        agent_col_width_frac=0.3,
        agent_col_padding_px=2,
        pad_grey=100,
    ):
        """Render the whole-grid human view"""

        if close:
            if self.window:
                self.window.close()
            return

        if mode == "human" and not self.window:
            self.window = SimpleImageViewer(caption="Marlgrid")

        # Compute which cells are visible to the agent
        highlight_mask = np.full((self.width, self.height), False, dtype=np.bool)
        for agent in self.agents:
            if agent.active:
                if self._full_obs:
                    xlow, ylow, xhigh, yhigh = 0, 0, self.width, self.height
                else:
                    xlow, ylow, xhigh, yhigh = agent.get_view_exts()

                dxlow, dylow = max(0, 0 - xlow), max(0, 0 - ylow)
                dxhigh, dyhigh = max(0, xhigh - self.grid.width), max(
                    0, yhigh - self.grid.height
                )
                if agent.see_through_walls:
                    highlight_mask[
                        xlow + dxlow : xhigh - dxhigh, ylow + dylow : yhigh - dyhigh
                    ] = True
                else:
                    a, b = self.gen_obs_grid(agent)
                    highlight_mask[
                        xlow + dxlow : xhigh - dxhigh, ylow + dylow : yhigh - dyhigh
                    ] |= rotate_grid(b, a.orientation)[
                        dxlow : (xhigh - xlow) - dxhigh, dylow : (yhigh - ylow) - dyhigh
                    ]

        # Render the whole grid
        img = self.grid.render(
            tile_size, highlight_mask=highlight_mask if highlight else None
        )
        rescale = lambda X, rescale_factor=2: np.kron(
            X, np.ones((int(rescale_factor), int(rescale_factor), 1))
        )

        if show_agent_views:

            target_partial_width = int(
                img.shape[0] * agent_col_width_frac - 2 * agent_col_padding_px
            )
            target_partial_height = (
                img.shape[1] - 2 * agent_col_padding_px
            ) // max_agents_per_col

            agent_views = [self.gen_agent_obs(agent) for agent in self.agents]
            agent_views = [
                view["pov"] if isinstance(view, dict) else view for view in agent_views
            ]
            agent_views = [
                rescale(
                    view,
                    min(
                        target_partial_width / view.shape[0],
                        target_partial_height / view.shape[1],
                    ),
                )
                for view in agent_views
            ]
            agent_views = [
                agent_views[pos : pos + max_agents_per_col]
                for pos in range(0, len(agent_views), max_agents_per_col)
            ]

            f_offset = (
                lambda view: np.array(
                    [
                        target_partial_height - view.shape[1],
                        target_partial_width - view.shape[0],
                    ]
                )
                // 2
            )

            cols = []
            for col_views in agent_views:
                col = np.full(
                    (img.shape[0], target_partial_width + 2 * agent_col_padding_px, 3),
                    pad_grey,
                    dtype=np.uint8,
                )
                for k, view in enumerate(col_views):
                    offset = f_offset(view) + agent_col_padding_px
                    offset[0] += k * target_partial_height
                    col[
                        offset[0] : offset[0] + view.shape[0],
                        offset[1] : offset[1] + view.shape[1],
                        :,
                    ] = view
                cols.append(col)

            img = np.concatenate((img, *cols), axis=1)

        if mode == "human":
            if not self.window.isopen:
                self.window.imshow(img)
                self.window.window.set_caption("Marlgrid")
            else:
                self.window.imshow(img)

        return img



import numpy as np
import torch
from torch import optim

from hive.utils.registry import registry
from hive.utils.utils import LossFn, OptimizerFn


def numpify(t):
    """Convert object to a numpy array.

    Args:
        t (np.ndarray | torch.Tensor | obj): Converts object to :py:class:`np.ndarray`.
    """
    if isinstance(t, np.ndarray):
        return t
    elif isinstance(t, torch.Tensor):
        return t.detach().cpu().numpy()
    else:
        return np.array(t)


class RMSpropTF(optim.Optimizer):
    """
    Direct cut-paste from rwhightman/pytorch-image-models.
    https://github.com/rwightman/pytorch-image-models/blob/f7d210d759beb00a3d0834a3ce2d93f6e17f3d38/timm/optim/rmsprop_tf.py
    Licensed under Apache 2.0, https://github.com/rwightman/pytorch-image-models/blob/master/LICENSE

    Implements RMSprop algorithm (TensorFlow style epsilon)

    NOTE: This is a direct cut-and-paste of PyTorch RMSprop with eps applied before sqrt
    and a few other modifications to closer match Tensorflow for matching hyper-params.
    Noteworthy changes include:

    1. Epsilon applied inside square-root
    2. square_avg initialized to ones
    3. LR scaling of update accumulated in momentum buffer

    Proposed by G. Hinton in his
    `course <http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf>`_.
    The centered version first appears in `Generating Sequences
    With Recurrent Neural Networks <https://arxiv.org/pdf/1308.0850v5.pdf>`_.

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-2)
        momentum (float, optional): momentum factor (default: 0)
        alpha (float, optional): smoothing (decay) constant (default: 0.9)
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-10)
        centered (bool, optional) : if ``True``, compute the centered RMSProp,
            the gradient is normalized by an estimation of its variance
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        decoupled_decay (bool, optional): decoupled weight decay as per https://arxiv.org/abs/1711.05101
        lr_in_momentum (bool, optional): learning rate scaling is included in the momentum buffer
            update as per defaults in Tensorflow

    """

    def __init__(
        self,
        params,
        lr=1e-2,
        alpha=0.9,
        eps=1e-10,
        weight_decay=0,
        momentum=0.0,
        centered=False,
        decoupled_decay=False,
        lr_in_momentum=True,
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= momentum:
            raise ValueError("Invalid momentum value: {}".format(momentum))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        if not 0.0 <= alpha:
            raise ValueError("Invalid alpha value: {}".format(alpha))

        defaults = dict(
            lr=lr,
            momentum=momentum,
            alpha=alpha,
            eps=eps,
            centered=centered,
            weight_decay=weight_decay,
            decoupled_decay=decoupled_decay,
            lr_in_momentum=lr_in_momentum,
        )
        super(RMSpropTF, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(RMSpropTF, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault("momentum", 0)
            group.setdefault("centered", False)

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad
                if grad.is_sparse:
                    raise RuntimeError("RMSprop does not support sparse gradients")
                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    state["square_avg"] = torch.ones_like(p)  # PyTorch inits to zero
                    if group["momentum"] > 0:
                        state["momentum_buffer"] = torch.zeros_like(p)
                    if group["centered"]:
                        state["grad_avg"] = torch.zeros_like(p)

                square_avg = state["square_avg"]
                one_minus_alpha = 1.0 - group["alpha"]

                state["step"] += 1

                if group["weight_decay"] != 0:
                    if group["decoupled_decay"]:
                        p.mul_(1.0 - group["lr"] * group["weight_decay"])
                    else:
                        grad = grad.add(p, alpha=group["weight_decay"])

                # Tensorflow order of ops for updating squared avg
                square_avg.add_(grad.pow(2) - square_avg, alpha=one_minus_alpha)
                # square_avg.mul_(alpha).addcmul_(grad, grad, value=1 - alpha)  # PyTorch original

                if group["centered"]:
                    grad_avg = state["grad_avg"]
                    grad_avg.add_(grad - grad_avg, alpha=one_minus_alpha)
                    avg = (
                        square_avg.addcmul(grad_avg, grad_avg, value=-1)
                        .add(group["eps"])
                        .sqrt_()
                    )  # eps in sqrt
                    # grad_avg.mul_(alpha).add_(grad, alpha=1 - alpha)  # PyTorch original
                else:
                    avg = square_avg.add(group["eps"]).sqrt_()  # eps moved in sqrt

                if group["momentum"] > 0:
                    buf = state["momentum_buffer"]
                    # Tensorflow accumulates the LR scaling in the momentum buffer
                    if group["lr_in_momentum"]:
                        buf.mul_(group["momentum"]).addcdiv_(
                            grad, avg, value=group["lr"]
                        )
                        p.add_(-buf)
                    else:
                        # PyTorch scales the param update by LR
                        buf.mul_(group["momentum"]).addcdiv_(grad, avg)
                        p.add_(buf, alpha=-group["lr"])
                else:
                    p.addcdiv_(grad, avg, value=-group["lr"])

        return loss


registry.register_all(
    OptimizerFn,
    {
        "Adadelta": OptimizerFn(optim.Adadelta),
        "Adagrad": OptimizerFn(optim.Adagrad),
        "Adam": OptimizerFn(optim.Adam),
        "Adamax": OptimizerFn(optim.Adamax),
        "AdamW": OptimizerFn(optim.AdamW),
        "ASGD": OptimizerFn(optim.ASGD),
        "LBFGS": OptimizerFn(optim.LBFGS),
        "RMSprop": OptimizerFn(optim.RMSprop),
        "RMSpropTF": OptimizerFn(RMSpropTF),
        "Rprop": OptimizerFn(optim.Rprop),
        "SGD": OptimizerFn(optim.SGD),
        "SparseAdam": OptimizerFn(optim.SparseAdam),
    },
)

registry.register_all(
    LossFn,
    {
        "BCELoss": LossFn(torch.nn.BCELoss),
        "BCEWithLogitsLoss": LossFn(torch.nn.BCEWithLogitsLoss),
        "CosineEmbeddingLoss": LossFn(torch.nn.CosineEmbeddingLoss),
        "CrossEntropyLoss": LossFn(torch.nn.CrossEntropyLoss),
        "CTCLoss": LossFn(torch.nn.CTCLoss),
        "HingeEmbeddingLoss": LossFn(torch.nn.HingeEmbeddingLoss),
        "KLDivLoss": LossFn(torch.nn.KLDivLoss),
        "L1Loss": LossFn(torch.nn.L1Loss),
        "MarginRankingLoss": LossFn(torch.nn.MarginRankingLoss),
        "MSELoss": LossFn(torch.nn.MSELoss),
        "MultiLabelMarginLoss": LossFn(torch.nn.MultiLabelMarginLoss),
        "MultiLabelSoftMarginLoss": LossFn(torch.nn.MultiLabelSoftMarginLoss),
        "MultiMarginLoss": LossFn(torch.nn.MultiMarginLoss),
        "NLLLoss": LossFn(torch.nn.NLLLoss),
        "NLLLoss2d": LossFn(torch.nn.NLLLoss2d),
        "PoissonNLLLoss": LossFn(torch.nn.PoissonNLLLoss),
        "SmoothL1Loss": LossFn(torch.nn.SmoothL1Loss),
        "SoftMarginLoss": LossFn(torch.nn.SoftMarginLoss),
        "TripletMarginLoss": LossFn(torch.nn.TripletMarginLoss),
    },
)

get_optimizer_fn = getattr(registry, f"get_{OptimizerFn.type_name()}")
get_loss_fn = getattr(registry, f"get_{LossFn.type_name()}")



import copy
import os

import numpy as np
import torch

from hive.agents.agent import Agent
from hive.agents.qnets.base import FunctionApproximator
from hive.agents.qnets.qnet_heads import DQNNetwork
from hive.agents.qnets.utils import (
    InitializationFn,
    calculate_output_dim,
    create_init_weights_fn,
)
from hive.replays import BaseReplayBuffer, CircularReplayBuffer
from hive.utils.loggers import Logger, NullLogger
from hive.utils.schedule import (
    LinearSchedule,
    PeriodicSchedule,
    Schedule,
    SwitchSchedule,
)
from hive.utils.utils import LossFn, OptimizerFn, create_folder, seeder


class DQNAgent(Agent):
    """An agent implementing the DQN algorithm. Uses an epsilon greedy
    exploration policy
    """

    def __init__(
        self,
        representation_net: FunctionApproximator,
        obs_dim,
        act_dim: int,
        id=0,
        optimizer_fn: OptimizerFn = None,
        loss_fn: LossFn = None,
        init_fn: InitializationFn = None,
        replay_buffer: BaseReplayBuffer = None,
        discount_rate: float = 0.99,
        n_step: int = 1,
        grad_clip: float = None,
        reward_clip: float = None,
        update_period_schedule: Schedule = None,
        target_net_soft_update: bool = False,
        target_net_update_fraction: float = 0.05,
        target_net_update_schedule: Schedule = None,
        epsilon_schedule: Schedule = None,
        test_epsilon: float = 0.001,
        min_replay_history: int = 5000,
        batch_size: int = 32,
        device="cpu",
        logger: Logger = None,
        log_frequency: int = 100,
    ):
        """
        Args:
            representation_net (FunctionApproximator): A network that outputs the
                representations that will be used to compute Q-values (e.g.
                everything except the final layer of the DQN).
            obs_dim: The shape of the observations.
            act_dim (int): The number of actions available to the agent.
            id: Agent identifier.
            optimizer_fn (OptimizerFn): A function that takes in a list of parameters
                to optimize and returns the optimizer. If None, defaults to
                :py:class:`~torch.optim.Adam`.
            loss_fn (LossFn): Loss function used by the agent. If None, defaults to
                :py:class:`~torch.nn.SmoothL1Loss`.
            init_fn (InitializationFn): Initializes the weights of qnet using
                create_init_weights_fn.
            replay_buffer (BaseReplayBuffer): The replay buffer that the agent will
                push observations to and sample from during learning. If None,
                defaults to
                :py:class:`~hive.replays.circular_replay.CircularReplayBuffer`.
            discount_rate (float): A number between 0 and 1 specifying how much
                future rewards are discounted by the agent.
            n_step (int): The horizon used in n-step returns to compute TD(n) targets.
            grad_clip (float): Gradients will be clipped to between
                [-grad_clip, grad_clip].
            reward_clip (float): Rewards will be clipped to between
                [-reward_clip, reward_clip].
            update_period_schedule (Schedule): Schedule determining how frequently
                the agent's Q-network is updated.
            target_net_soft_update (bool): Whether the target net parameters are
                replaced by the qnet parameters completely or using a weighted
                average of the target net parameters and the qnet parameters.
            target_net_update_fraction (float): The weight given to the target
                net parameters in a soft update.
            target_net_update_schedule (Schedule): Schedule determining how frequently
                the target net is updated.
            epsilon_schedule (Schedule): Schedule determining the value of epsilon
                through the course of training.
            test_epsilon (float): epsilon (probability of choosing a random action)
                to be used during testing phase.
            min_replay_history (int): How many observations to fill the replay buffer
                with before starting to learn.
            batch_size (int): The size of the batch sampled from the replay buffer
                during learning.
            device: Device on which all computations should be run.
            logger (ScheduledLogger): Logger used to log agent's metrics.
            log_frequency (int): How often to log the agent's metrics.
        """
        super().__init__(obs_dim=obs_dim, act_dim=act_dim, id=id)
        self._init_fn = create_init_weights_fn(init_fn)
        self._device = torch.device("cpu" if not torch.cuda.is_available() else device)
        self.create_q_networks(representation_net)
        if optimizer_fn is None:
            optimizer_fn = torch.optim.Adam
        self._optimizer = optimizer_fn(self._qnet.parameters())
        self._rng = np.random.default_rng(seed=seeder.get_new_seed())
        self._replay_buffer = replay_buffer
        if self._replay_buffer is None:
            self._replay_buffer = CircularReplayBuffer()
        self._discount_rate = discount_rate ** n_step
        self._grad_clip = grad_clip
        self._reward_clip = reward_clip
        self._target_net_soft_update = target_net_soft_update
        self._target_net_update_fraction = target_net_update_fraction
        if loss_fn is None:
            loss_fn = torch.nn.SmoothL1Loss
        self._loss_fn = loss_fn(reduction="none")
        self._batch_size = batch_size
        self._logger = logger
        if self._logger is None:
            self._logger = NullLogger([])
        self._timescale = self.id
        self._logger.register_timescale(
            self._timescale, PeriodicSchedule(False, True, log_frequency)
        )
        self._update_period_schedule = update_period_schedule
        if self._update_period_schedule is None:
            self._update_period_schedule = PeriodicSchedule(False, True, 1)
        self._target_net_update_schedule = target_net_update_schedule
        if self._target_net_update_schedule is None:
            self._target_net_update_schedule = PeriodicSchedule(False, True, 10000)
        self._epsilon_schedule = epsilon_schedule
        if self._epsilon_schedule is None:
            self._epsilon_schedule = LinearSchedule(1, 0.1, 100000)
        self._test_epsilon = test_epsilon
        self._learn_schedule = SwitchSchedule(False, True, min_replay_history)

        self._state = {"episode_start": True}
        self._training = False

    def create_q_networks(self, representation_net):
        """Creates the Q-network and target Q-network.

        Args:
            representation_net: A network that outputs the representations that will
                be used to compute Q-values (e.g. everything except the final layer
                of the DQN).
        """
        network = representation_net(self._obs_dim)
        network_output_dim = np.prod(calculate_output_dim(network, self._obs_dim))
        self._qnet = DQNNetwork(network, network_output_dim, self._act_dim).to(
            self._device
        )
        self._qnet.apply(self._init_fn)
        self._target_qnet = copy.deepcopy(self._qnet).requires_grad_(False)

    def train(self):
        """Changes the agent to training mode."""
        super().train()
        self._qnet.train()
        self._target_qnet.train()

    def eval(self):
        """Changes the agent to evaluation mode."""
        super().eval()
        self._qnet.eval()
        self._target_qnet.eval()

    def preprocess_update_info(self, update_info):
        """Preprocesses the :obj:`update_info` before it goes into the replay buffer.
        Clips the reward in update_info.

        Args:
            update_info: Contains the information from the current timestep that the
                agent should use to update itself.
        """
        if self._reward_clip is not None:
            update_info["reward"] = np.clip(
                update_info["reward"], -self._reward_clip, self._reward_clip
            )
        preprocessed_update_info = {
            "observation": update_info["observation"],
            "action": update_info["action"],
            "reward": update_info["reward"],
            "done": update_info["done"],
        }
        if "agent_id" in update_info:
            preprocessed_update_info["agent_id"] = int(update_info["agent_id"])

        return preprocessed_update_info

    def preprocess_update_batch(self, batch):
        """Preprocess the batch sampled from the replay buffer.

        Args:
            batch: Batch sampled from the replay buffer for the current update.

        Returns:
            (tuple):
                - (tuple) Inputs used to calculate current state values.
                - (tuple) Inputs used to calculate next state values
                - Preprocessed batch.
        """
        for key in batch:
            batch[key] = torch.tensor(batch[key], device=self._device)
        return (batch["observation"],), (batch["next_observation"],), batch

    @torch.no_grad()
    def act(self, observation):
        """Returns the action for the agent. If in training mode, follows an epsilon
        greedy policy. Otherwise, returns the action with the highest Q-value.

        Args:
            observation: The current observation.
        """

        # Determine and log the value of epsilon
        if self._training:
            if not self._learn_schedule.get_value():
                epsilon = 1.0
            else:
                epsilon = self._epsilon_schedule.update()
            if self._logger.update_step(self._timescale):
                self._logger.log_scalar("epsilon", epsilon, self._timescale)
        else:
            epsilon = self._test_epsilon

        # Sample action. With epsilon probability choose random action,
        # otherwise select the action with the highest q-value.
        observation = torch.tensor(
            np.expand_dims(observation, axis=0), device=self._device
        ).float()
        qvals = self._qnet(observation)
        if self._rng.random() < epsilon:
            action = self._rng.integers(self._act_dim)
        else:
            # Note: not explicitly handling the ties
            action = torch.argmax(qvals).item()

        if (
            self._training
            and self._logger.should_log(self._timescale)
            and self._state["episode_start"]
        ):
            self._logger.log_scalar("train_qval", torch.max(qvals), self._timescale)
            self._state["episode_start"] = False
        return action

    def update(self, update_info):
        """
        Updates the DQN agent.

        Args:
            update_info: dictionary containing all the necessary information to
                update the agent. Should contain a full transition, with keys for
                "observation", "action", "reward", and "done".
        """
        if update_info["done"]:
            self._state["episode_start"] = True

        if not self._training:
            return

        # Add the most recent transition to the replay buffer.
        self._replay_buffer.add(**self.preprocess_update_info(update_info))

        # Update the q network based on a sample batch from the replay buffer.
        # If the replay buffer doesn't have enough samples, catch the exception
        # and move on.
        if (
            self._learn_schedule.update()
            and self._replay_buffer.size() > 0
            and self._update_period_schedule.update()
        ):
            batch = self._replay_buffer.sample(batch_size=self._batch_size)
            (
                current_state_inputs,
                next_state_inputs,
                batch,
            ) = self.preprocess_update_batch(batch)

            # Compute predicted Q values
            self._optimizer.zero_grad()
            pred_qvals = self._qnet(*current_state_inputs)
            actions = batch["action"].long()
            pred_qvals = pred_qvals[torch.arange(pred_qvals.size(0)), actions]

            # Compute 1-step Q targets
            next_qvals = self._target_qnet(*next_state_inputs)
            next_qvals, _ = torch.max(next_qvals, dim=1)

            q_targets = batch["reward"] + self._discount_rate * next_qvals * (
                1 - batch["done"]
            )

            loss = self._loss_fn(pred_qvals, q_targets).mean()

            if self._logger.should_log(self._timescale):
                self._logger.log_scalar("train_loss", loss, self._timescale)

            loss.backward()
            if self._grad_clip is not None:
                torch.nn.utils.clip_grad_value_(
                    self._qnet.parameters(), self._grad_clip
                )
            self._optimizer.step()

        # Update target network
        if self._target_net_update_schedule.update():
            self._update_target()

    def _update_target(self):
        """Update the target network."""
        if self._target_net_soft_update:
            target_params = self._target_qnet.state_dict()
            current_params = self._qnet.state_dict()
            for key in list(target_params.keys()):
                target_params[key] = (
                    1 - self._target_net_update_fraction
                ) * target_params[
                    key
                ] + self._target_net_update_fraction * current_params[
                    key
                ]
            self._target_qnet.load_state_dict(target_params)
        else:
            self._target_qnet.load_state_dict(self._qnet.state_dict())

    def save(self, dname):
        torch.save(
            {
                "qnet": self._qnet.state_dict(),
                "target_qnet": self._target_qnet.state_dict(),
                "optimizer": self._optimizer.state_dict(),
                "learn_schedule": self._learn_schedule,
                "epsilon_schedule": self._epsilon_schedule,
                "target_net_update_schedule": self._target_net_update_schedule,
                "rng": self._rng,
            },
            os.path.join(dname, "agent.pt"),
        )
        replay_dir = os.path.join(dname, "replay")
        create_folder(replay_dir)
        self._replay_buffer.save(replay_dir)

    def load(self, dname):
        checkpoint = torch.load(os.path.join(dname, "agent.pt"))
        self._qnet.load_state_dict(checkpoint["qnet"])
        self._target_qnet.load_state_dict(checkpoint["target_qnet"])
        self._optimizer.load_state_dict(checkpoint["optimizer"])
        self._learn_schedule = checkpoint["learn_schedule"]
        self._epsilon_schedule = checkpoint["epsilon_schedule"]
        self._target_net_update_schedule = checkpoint["target_net_update_schedule"]
        self._rng = checkpoint["rng"]
        self._replay_buffer.load(os.path.join(dname, "replay"))



from abc import ABC, abstractmethod

from hive.utils.registry import Registrable


class BaseEnv(ABC, Registrable):
    """
    Base class for environments.
    """

    def __init__(self, env_spec, num_players):
        """
        Args:
            env_spec (EnvSpec): An object containing information about the
                environment.
            num_players (int): The number of players in the environment.
        """
        self._env_spec = env_spec
        self._num_players = num_players
        self._turn = 0

    @abstractmethod
    def reset(self):
        """
        Resets the state of the environment.

        Returns:
            observation: The initial observation of the new episode.
            turn (int): The index of the agent which should take turn.
        """
        raise NotImplementedError

    @abstractmethod
    def step(self, action):
        """
        Run one time-step of the environment using the input action.

        Args:
            action: An element of environment's action space.

        Returns:
            observation: Indicates the next state that is an element of environment's observation space.
            reward: A reward achieved from the transition.
            done (bool): Indicates whether the episode has ended.
            turn (int): Indicates which agent should take turn.
            info (dict): Additional custom information.
        """

        raise NotImplementedError

    def render(self, mode="rgb_array"):
        """
        Displays a rendered frame from the environment.
        """
        raise NotImplementedError

    @abstractmethod
    def seed(self, seed=None):
        """
        Reseeds the environment.

        Args:
            seed (int): Seed to use for environment.
        """
        raise NotImplementedError

    def save(self, save_dir):
        """
        Saves the environment.

        Args:
            save_dir (str): Location to save environment state.
        """
        raise NotImplementedError

    def load(self, load_dir):
        """
        Loads the environment.

        Args:
            load_dir (str): Location to load environment state from.
        """
        raise NotImplementedError

    def close(self):
        """
        Additional clean up operations
        """
        raise NotImplementedError

    @property
    def env_spec(self):
        return self._env_spec

    @env_spec.setter
    def env_spec(self, env_spec):
        self._env_spec = env_spec

    @classmethod
    def type_name(cls):
        """
        Returns: "env"
        """
        return "env"


class ParallelEnv(BaseEnv):
    """Base class for environments that take make all agents step in parallel.

    ParallelEnv takes an environment that expects an array of actions at each step
    to execute in parallel, and allows you to instead pass it a single action at each
    step.

    This class makes use of Python's multiple inheritance pattern. Specifically,
    when writing your parallel environment, it should extend both this class and
    the class that implements the step method that takes in actions for all agents.

    If environment class A has the logic for the step function that takes in the
    array of actions, and environment class B is your parallel step version of that
    environment, class B should be defined as:

    .. code-block:: python

        class B(ParallelEnv, A):
            ...

    The order in which you list the classes is important. ParallelEnv **must** come
    before A in the order.
    """

    def __init__(self, env_name, num_players):
        super().__init__(env_name, num_players)
        self._actions = []
        self._obs = None
        self._info = None
        self._done = False

    def reset(self):
        self._obs, _ = super().reset()
        return self._obs[0], 0

    def step(self, action):
        self._actions.append(action)
        if len(self._actions) == self._num_players:
            observation, reward, done, _, info = super().step(self._actions)
            self._actions = []
            self._turn = 0
            self._obs = observation
            self._info = info
            self._done = done
        else:
            self._turn = (self._turn + 1) % self._num_players
            reward = 0
        return (
            self._obs[self._turn],
            reward,
            self._done and self._turn == 0,
            self._turn,
            self._info,
        )



import torch
import torch.nn.functional as F
from torch import nn


class DQNNetwork(nn.Module):
    """Implements the standard DQN value computation. Transforms output from
    :obj:`base_network` with output dimension :obj:`hidden_dim` to dimension
    :obj:`out_dim`, which should be equal to the number of actions.
    """

    def __init__(
        self,
        base_network: nn.Module,
        hidden_dim: int,
        out_dim: int,
        linear_fn: nn.Module = None,
    ):
        """
        Args:
            base_network (torch.nn.Module): Backbone network that computes the
                representations that are used to compute action values.
            hidden_dim (int): Dimension of the output of the :obj:`network`.
            out_dim (int): Output dimension of the DQN. Should be equal to the
                number of actions that you are computing values for.
            linear_fn (torch.nn.Module): Function that will create the
                :py:class:`torch.nn.Module` that will take the output of
                :obj:`network` and produce the final action values. If
                :obj:`None`, a :py:class:`torch.nn.Linear` layer will be used.
        """
        super().__init__()
        self.base_network = base_network
        self._linear_fn = linear_fn if linear_fn is not None else nn.Linear
        self.output_layer = self._linear_fn(hidden_dim, out_dim)

    def forward(self, x):
        x = self.base_network(x)
        x = x.flatten(start_dim=1)
        return self.output_layer(x)


class DuelingNetwork(nn.Module):
    """Computes action values using Dueling Networks (https://arxiv.org/abs/1511.06581).
    In dueling, we have two heads---one for estimating advantage function and one for
    estimating value function.
    """

    def __init__(
        self,
        base_network: nn.Module,
        hidden_dim: int,
        out_dim: int,
        linear_fn: nn.Module = None,
        atoms: int = 1,
    ):
        """
        Args:
            base_network (torch.nn.Module): Backbone network that computes the
                representations that are shared by the two estimators.
            hidden_dim (int): Dimension of the output of the :obj:`base_network`.
            out_dim (int): Output dimension of the Dueling DQN. Should be equal
                to the number of actions that you are computing values for.
            linear_fn (torch.nn.Module): Function that will create the
                :py:class:`torch.nn.Module` that will take the output of
                :obj:`network` and produce the final action values. If
                :obj:`None`, a :py:class:`torch.nn.Linear` layer will be used.
            atoms (int): Multiplier for the dimension of the output. For standard
                dueling networks, this should be 1. Used by
                :py:class:`~hive.agents.qnets.qnet_heads.DistributionalNetwork`.
        """
        super().__init__()
        self.base_network = base_network
        self._hidden_dim = hidden_dim
        self._out_dim = out_dim
        self._atoms = atoms
        self._linear_fn = linear_fn if linear_fn is not None else nn.Linear
        self.init_networks()

    def init_networks(self):
        self.output_layer_adv = self._linear_fn(
            self._hidden_dim, self._out_dim * self._atoms
        )

        self.output_layer_val = self._linear_fn(self._hidden_dim, 1 * self._atoms)

    def forward(self, x):
        x = self.base_network(x)
        x = x.flatten(start_dim=1)
        adv = self.output_layer_adv(x)
        val = self.output_layer_val(x)

        if adv.dim() == 1:
            x = val + adv - adv.mean(0)
        else:
            adv = adv.reshape(adv.size(0), self._out_dim, self._atoms)
            val = val.reshape(val.size(0), 1, self._atoms)
            x = val + adv - adv.mean(dim=1, keepdim=True)
            if self._atoms == 1:
                x = x.squeeze(dim=2)
        return x


class DistributionalNetwork(nn.Module):
    """Computes a categorical distribution over values for each action
    (https://arxiv.org/abs/1707.06887)."""

    def __init__(
        self,
        base_network: nn.Module,
        out_dim: int,
        vmin: float = 0,
        vmax: float = 200,
        atoms: int = 51,
    ):
        """
        Args:
            base_network (torch.nn.Module): Backbone network that computes the
                representations that are used to compute the value distribution.
            out_dim (int): Output dimension of the Distributional DQN. Should be
                equal to the number of actions that you are computing values for.
            vmin (float): The minimum of the support of the categorical value
                distribution.
            vmax (float): The maximum of the support of the categorical value
                distribution.
            atoms (int): Number of atoms discretizing the support range of the
                categorical value distribution.
        """

        super().__init__()
        self.base_network = base_network
        self._supports = torch.nn.Parameter(torch.linspace(vmin, vmax, atoms))
        self._out_dim = out_dim
        self._atoms = atoms

    def forward(self, x):
        x = self.dist(x)
        x = torch.sum(x * self._supports, dim=2)
        return x

    def dist(self, x):
        """Computes a categorical distribution over values for each action."""
        x = self.base_network(x)
        x = x.view(-1, self._out_dim, self._atoms)
        x = F.softmax(x, dim=-1)
        return x



import abc
import copy
import os
from typing import List

import torch
import wandb

from hive.utils.registry import Registrable, registry
from hive.utils.schedule import ConstantSchedule, Schedule, get_schedule
from hive.utils.utils import Chomp, create_folder


class Logger(abc.ABC, Registrable):
    """Abstract class for logging in hive."""

    def __init__(self, timescales=None):
        """Constructor for base Logger class. Every Logger must call this constructor
        in its own constructor

        Args:
            timescales (str | list(str)): The different timescales at which logger
                needs to log. If only logging at one timescale, it is acceptable to
                only pass a string.
        """
        if timescales is None:
            self._timescales = []
        elif isinstance(timescales, str):
            self._timescales = [timescales]
        elif isinstance(timescales, list):
            self._timescales = timescales
        else:
            raise ValueError("Need string or list of strings for timescales")

    def register_timescale(self, timescale):
        """Register a new timescale with the logger.

        Args:
            timescale (str): Timescale to register.
        """
        self._timescales.append(timescale)

    @abc.abstractmethod
    def log_config(self, config):
        """Log the config.

        Args:
            config (dict): Config parameters.
        """
        pass

    @abc.abstractmethod
    def log_scalar(self, name, value, prefix):
        """Log a scalar variable.

        Args:
            name (str): Name of the metric to be logged.
            value (float): Value to be logged.
            prefix (str): Prefix to append to metric name.
        """
        pass

    @abc.abstractmethod
    def log_metrics(self, metrics, prefix):
        """Log a dictionary of values.

        Args:
            metrics (dict): Dictionary of metrics to be logged.
            prefix (str): Prefix to append to metric name.
        """
        pass

    @abc.abstractmethod
    def save(self, dir_name):
        """Saves the current state of the log files.

        Args:
            dir_name (str): Name of the directory to save the log files.
        """
        pass

    @abc.abstractmethod
    def load(self, dir_name):
        """Loads the log files from given directory.

        Args:
            dir_name (str): Name of the directory to load the log file from.
        """
        pass

    @classmethod
    def type_name(cls):
        return "logger"


class ScheduledLogger(Logger):
    """Abstract class that manages a schedule for logging.

    The update_step method should be called for each step in the loop to update
    the logger's schedule. The should_log method can be used to check whether
    the logger should log anything.

    This schedule is not strictly enforced! It is still possible to log something
    even if should_log returns false. These functions are just for the purpose
    of convenience.
    """

    def __init__(self, timescales=None, logger_schedules=None):
        """
        Any timescales not assigned schedule from logger_schedules will be assigned
        a ConstantSchedule(True).

        Args:
            timescales (str|list[str]): The different timescales at which logger needs
                to log. If only logging at one timescale, it is acceptable to only pass
                a string.
            logger_schedules (Schedule|list|dict): Schedules used to keep track of when
                to log. If a single schedule, it is copied for each timescale. If a
                list of schedules, the schedules are matched up in order with the list
                of timescales provided. If a dictionary, the keys should be the
                timescale and the values should be the schedule.
        """
        super().__init__(timescales)
        if logger_schedules is None:
            logger_schedules = ConstantSchedule(True)
        if isinstance(logger_schedules, dict):
            self._logger_schedules = logger_schedules
        elif isinstance(logger_schedules, list):
            self._logger_schedules = {
                self._timescales[idx]: logger_schedules[idx]
                for idx in range(min(len(logger_schedules), len(self._timescales)))
            }
        elif isinstance(logger_schedules, Schedule):
            self._logger_schedules = {
                timescale: copy.deepcopy(logger_schedules)
                for timescale in self._timescales
            }
        else:
            raise ValueError(
                "logger_schedule must be a dict, list of Schedules, or Schedule object"
            )
        for timescale, schedule in self._logger_schedules.items():
            if isinstance(schedule, dict):
                self._logger_schedules[timescale] = get_schedule(
                    schedule["name"], schedule["kwargs"]
                )

        for timescale in self._timescales:
            if timescale not in self._logger_schedules:
                self._logger_schedules[timescale] = ConstantSchedule(True)
        self._steps = {timescale: 0 for timescale in self._timescales}

    def register_timescale(self, timescale, schedule=None):
        """Register a new timescale.

        Args:
            timescale (str): Timescale to register.
            schedule (Schedule): Schedule to use for this timescale.
        """
        super().register_timescale(timescale)
        if schedule is None:
            schedule = ConstantSchedule(True)
        self._logger_schedules[timescale] = schedule
        self._steps[timescale] = 0

    def update_step(self, timescale):
        """Update the step and schedule for a given timescale.

        Args:
            timescale (str): A registered timescale.
        """
        self._steps[timescale] += 1
        self._logger_schedules[timescale].update()
        return self.should_log(timescale)

    def should_log(self, timescale):
        """Check if you should log for a given timescale.

        Args:
            timescale (str): A registered timescale.
        """
        return self._logger_schedules[timescale].get_value()

    def save(self, dir_name):
        logger_state = Chomp()
        logger_state.timescales = self._timescales
        logger_state.schedules = self._logger_schedules
        logger_state.steps = self._steps
        logger_state.save(os.path.join(dir_name, "logger_state.p"))

    def load(self, dir_name):
        logger_state = Chomp()
        logger_state.load(os.path.join(dir_name, "logger_state.p"))
        self._timescales = logger_state.timescales
        self._logger_schedules = logger_state.schedules
        self._steps = logger_state.steps


class NullLogger(ScheduledLogger):
    """A null logger that does not log anything.

    Used if you don't want to log anything, but still want to use parts of the
    framework that ask for a logger.
    """

    def __init__(self, timescales=None, logger_schedules=None):
        super().__init__(timescales, logger_schedules)

    def log_config(self, config):
        pass

    def log_scalar(self, name, value, timescale):
        pass

    def log_metrics(self, metrics, timescale):
        pass

    def save(self, dir_name):
        pass

    def load(self, dir_name):
        pass


class WandbLogger(ScheduledLogger):
    """A Wandb logger.

    This logger can be used to log to wandb. It assumes that wandb is configured
    locally on your system. Multiple timescales/loggers can be implemented by
    instantiating multiple loggers with different logger_names. These should still
    have the same project and run names.

    Check the wandb documentation for more details on the parameters.
    """

    def __init__(
        self,
        timescales=None,
        logger_schedules=None,
        project=None,
        name=None,
        dir=None,
        mode=None,
        id=None,
        resume=None,
        start_method=None,
        **kwargs,
    ):
        """
        Args:
            timescales (str|list[str]): The different timescales at which logger needs
                to log. If only logging at one timescale, it is acceptable to only pass
                a string.
            logger_schedules (Schedule|list|dict): Schedules used to keep track of when
                to log. If a single schedule, it is copied for each timescale. If a
                list of schedules, the schedules are matched up in order with the list
                of timescales provided. If a dictionary, the keys should be the
                timescale and the values should be the schedule.
            project (str): Name of the project. Wandb's dash groups all runs with
                the same project name together.
            name (str): Name of the run. Used to identify the run on the wandb
                dash.
            dir (str): Local directory where wandb saves logs.
            mode (str): The mode of logging. Can be "online", "offline" or "disabled".
                In offline mode, writes all data to disk for later syncing to a server,
                while in disabled mode, it makes all calls to wandb api's noop's, while
                maintaining core functionality.
            id (str, optional): A unique ID for this run, used for resuming.
                It must be unique in the project, and if you delete a run you can't
                reuse the ID.
            resume (bool, str, optional): Sets the resuming behavior.
                Options are the same as mentioned in Wandb's doc.
            start_method (str): The start method to use for wandb's process. See
                https://docs.wandb.ai/guides/track/launch#init-start-error.
            **kwargs: You can pass any other arguments to wandb's init method as
                keyword arguments. Note, these arguments can't be overriden from the
                command line.
        """
        super().__init__(timescales, logger_schedules)

        settings = None
        if start_method is not None:
            settings = wandb.Settings(start_method=start_method)

        wandb.init(
            project=project,
            name=name,
            dir=dir,
            mode=mode,
            id=id,
            resume=resume,
            settings=settings,
            **kwargs,
        )

    def log_config(self, config):
        # Convert list parameters to nested dictionary
        for k, v in config.items():
            if isinstance(v, list):
                config[k] = {}
                for idx, param in enumerate(v):
                    config[k][idx] = param

        wandb.config.update(config)

    def log_scalar(self, name, value, prefix):
        metrics = {f"{prefix}/{name}": value}
        metrics.update(
            {
                f"{timescale}_step": self._steps[timescale]
                for timescale in self._timescales
            }
        )
        wandb.log(metrics)

    def log_metrics(self, metrics, prefix):
        metrics = {f"{prefix}/{name}": value for (name, value) in metrics.items()}
        metrics.update(
            {
                f"{timescale}_step": self._steps[timescale]
                for timescale in self._timescales
            }
        )
        wandb.log(metrics)


class ChompLogger(ScheduledLogger):
    """This logger uses the Chomp data structure to store all logged values which are
    then directly saved to disk.
    """

    def __init__(self, timescales=None, logger_schedules=None):
        super().__init__(timescales, logger_schedules)
        self._log_data = Chomp()

    def log_config(self, config):
        self._log_data["config"] = config

    def log_scalar(self, name, value, prefix):
        metric_name = f"{prefix}/{name}"
        if metric_name not in self._log_data:
            self._log_data[metric_name] = [[], []]
        if isinstance(value, torch.Tensor):
            self._log_data[metric_name][0].append(value.item())
        else:
            self._log_data[metric_name][0].append(value)
        self._log_data[metric_name][1].append(
            {timescale: self._steps[timescale] for timescale in self._timescales}
        )

    def log_metrics(self, metrics, prefix):
        for name in metrics:
            metric_name = f"{prefix}/{name}"
            if metric_name not in self._log_data:
                self._log_data[metric_name] = [[], []]
            if isinstance(metrics[name], torch.Tensor):
                self._log_data[metric_name][0].append(metrics[name].item())
            else:
                self._log_data[metric_name][0].append(metrics[name])
            self._log_data[metric_name][1].append(
                {timescale: self._steps[timescale] for timescale in self._timescales}
            )

    def save(self, dir_name):
        super().save(dir_name)
        self._log_data.save(os.path.join(dir_name, "log_data.p"))

    def load(self, dir_name):
        super().load(dir_name)
        self._log_data.load(os.path.join(dir_name, "log_data.p"))


class CompositeLogger(Logger):
    """This Logger aggregates multiple loggers together.

    This logger is for convenience and allows for logging using multiple loggers without
    having to keep track of several loggers. When timescales are updated, this logger
    updates the timescale for each one of its component loggers. When logging, logs to
    each of its component loggers as long as the logger is not a ScheduledLogger that
    should not be logging for the timescale.
    """

    def __init__(self, logger_list: List[Logger]):
        super().__init__([])
        self._logger_list = logger_list

    def register_timescale(self, timescale, schedule=None):
        for logger in self._logger_list:
            if isinstance(logger, ScheduledLogger):
                logger.register_timescale(timescale, schedule)
            else:
                logger.register_timescale(timescale)

    def log_config(self, config):
        for logger in self._logger_list:
            logger.log_config(config)

    def log_scalar(self, name, value, prefix):
        for logger in self._logger_list:
            logger.log_scalar(name, value, prefix)

    def log_metrics(self, metrics, prefix):
        for logger in self._logger_list:
            logger.log_metrics(metrics, prefix=prefix)

    def update_step(self, timescale):
        """Update the step and schedule for a given timescale for every
        ScheduledLogger.

        Args:
            timescale (str): A registered timescale.
        """

        for logger in self._logger_list:
            if isinstance(logger, ScheduledLogger):
                logger.update_step(timescale)
        return self.should_log(timescale)

    def should_log(self, timescale):
        """Check if you should log for a given timescale. If any logger in the list
        is scheduled to log, returns True.

        Args:
            timescale (str): A registered timescale.
        """
        for logger in self._logger_list:
            if not isinstance(logger, ScheduledLogger) or logger.should_log(timescale):
                return True
        return False

    def save(self, dir_name):
        for idx, logger in enumerate(self._logger_list):
            save_dir = os.path.join(dir_name, f"logger_{idx}")
            create_folder(save_dir)
            logger.save(save_dir)

    def load(self, dir_name):
        for idx, logger in enumerate(self._logger_list):
            load_dir = os.path.join(dir_name, f"logger_{idx}")
            logger.load(load_dir)


registry.register_all(
    Logger,
    {
        "NullLogger": NullLogger,
        "WandbLogger": WandbLogger,
        "ChompLogger": ChompLogger,
        "CompositeLogger": CompositeLogger,
    },
)

get_logger = getattr(registry, f"get_{Logger.type_name()}")



import abc

from hive.utils.registry import Registrable


class Agent(abc.ABC, Registrable):
    """Base class for agents. Every implemented agent should be a subclass of
    this class.
    """

    def __init__(self, obs_dim, act_dim, id=0):
        """
        Args:
            obs_dim: Dimension of observations that agent will see.
            act_dim: Number of actions that the agent needs to chose from.
            id: Identifier for the agent.
        """
        self._obs_dim = obs_dim
        self._act_dim = act_dim
        self._training = True
        self._id = str(id)

    @property
    def id(self):
        return self._id

    @abc.abstractmethod
    def act(self, observation):
        """Returns an action for the agent to perform based on the observation.

        Args:
            observation: Current observation that agent should act on.
        Returns:
            Action for the current timestep.
        """
        pass

    @abc.abstractmethod
    def update(self, update_info):
        """
        Updates the agent.

        Args:
            update_info (dict): Contains information agent needs to update
                itself.
        """
        pass

    def train(self):
        """Changes the agent to training mode."""
        self._training = True

    def eval(self):
        """Changes the agent to evaluation mode"""
        self._training = False

    @abc.abstractmethod
    def save(self, dname):
        """
        Saves agent checkpointing information to file for future loading.

        Args:
            dname (str): directory where agent should save all relevant info.
        """
        pass

    @abc.abstractmethod
    def load(self, dname):
        """
        Loads agent information from file.

        Args:
            dname (str): directory where agent checkpoint info is stored.
        """
        pass

    @classmethod
    def type_name(cls):
        """
        Returns:
            "agent"
        """
        return "agent"



"""Implementation of a simple experiment class."""
import logging
import os

import yaml

from hive.utils.utils import Chomp, create_folder


class Experiment(object):
    """Implementation of a simple experiment class."""

    def __init__(self, name, dir_name, schedule):
        """Initializes an experiment object.

        The experiment state is an exposed property of objects of this class. It can
        be used to keep track of objects that need to be saved to keep track of the
        experiment, but don't fit in one of the standard categories. One example of
        this is the various schedules used in the Runner class.

        Args:
            name (str): Name of the experiment.
            dir_name (str): Absolute path to the directory to save/load the experiment.
        """

        self._name = name
        self._dir_name = os.path.join(dir_name, name)
        self._schedule = schedule
        self._step = 0
        create_folder(self._dir_name)

        self._config = None
        self._logger = None
        self._agents = None
        self._environment = None
        self.experiment_state = Chomp()
        self.experiment_state["saving_schedule"] = self._schedule

    def register_experiment(
        self,
        config=None,
        logger=None,
        agents=None,
        environment=None,
    ):
        """Registers all the components of an experiment.

        Args:
            config (Chomp): a config dictionary.
            logger (Logger): a logger object.
            agents (Agent | list[Agent]): either an agent object or a list of agents.
            environment (BaseEnv): an environment object.
        """

        self._config = config
        self._logger = logger
        self._logger.log_config(config)

        if agents is not None and not isinstance(agents, list):
            agents = [agents]
        self._agents = agents
        self._environment = environment

    def update_step(self):
        """Updates the step of the saving schedule for the experiment."""
        self._step += 1
        return self._schedule.update()

    def should_save(self):
        """Returns whether you should save the experiment at the current step."""
        return self._schedule.get_value()

    def save(self, tag="current"):
        """Saves the experiment.
        Args:
            tag (str): Tag to prefix the folder.
        """

        save_dir = os.path.join(self._dir_name, tag)
        create_folder(save_dir)

        logging.info("Saving the experiment at {}".format(save_dir))

        flag_file = os.path.join(save_dir, "flag.p")
        if os.path.isfile(flag_file):
            os.remove(flag_file)

        if self._config is not None:
            file_name = os.path.join(save_dir, "config.yml")
            with open(file_name, "w") as f:
                yaml.safe_dump(dict(self._config), f)

        if self._logger is not None:
            folder_name = os.path.join(save_dir, "logger")
            create_folder(folder_name)
            self._logger.save(folder_name)

        if self._agents is not None:
            for idx, agent in enumerate(self._agents):
                agent_dir = os.path.join(save_dir, f"agent_{idx}")
                create_folder(agent_dir)
                agent.save(agent_dir)

        if self._environment is not None:
            file_name = os.path.join(save_dir, "environment.p")
            self._environment.save(file_name)

        file_name = os.path.join(save_dir, "experiment_state.p")
        self.experiment_state.save(file_name)

        file = open(flag_file, "w")
        file.close()

    def is_resumable(self, tag="current"):
        """Returns true if the experiment is resumable.

        Args:
            tag (str): Tag for the saved experiment.
        """

        flag_file = os.path.join(self._dir_name, tag, "flag.p")
        if os.path.isfile(flag_file):
            return True
        else:
            return False

    def resume(self, tag="current"):
        """Resumes the experiment from a checkpoint.

        Args:
            tag (str): Tag for the saved experiment.
        """

        if self.is_resumable(tag):
            save_dir = os.path.join(self._dir_name, tag)
            logging.info("Loading the experiment from {}".format(save_dir))

            if self._config is not None:
                file_name = os.path.join(save_dir, "config.yml")
                with open(file_name) as f:
                    self._config = Chomp(yaml.safe_load(f))

            if self._logger is not None:
                folder_name = os.path.join(save_dir, "logger")
                self._logger.load(folder_name)

            if self._agents is not None:
                for idx, agent in enumerate(self._agents):
                    agent_dir = os.path.join(save_dir, f"agent_{idx}")
                    agent.load(agent_dir)

            if self._environment is not None:
                file_name = os.path.join(save_dir, "environment.p")
                self._environment.load(file_name)

            file_name = os.path.join(save_dir, "experiment_state.p")
            self.experiment_state.load(file_name)
            self._schedule = self.experiment_state["saving_schedule"]



from hive.replays.circular_replay import CircularReplayBuffer, SimpleReplayBuffer
from hive.replays.legal_moves_replay import LegalMovesBuffer
from hive.replays.prioritized_replay import PrioritizedReplayBuffer
from hive.replays.replay_buffer import BaseReplayBuffer
from hive.utils.registry import registry

registry.register_all(
    BaseReplayBuffer,
    {
        "CircularReplayBuffer": CircularReplayBuffer,
        "SimpleReplayBuffer": SimpleReplayBuffer,
        "PrioritizedReplayBuffer": PrioritizedReplayBuffer,
        "LegalMovesBuffer": LegalMovesBuffer,
    },
)

get_replay = getattr(registry, f"get_{BaseReplayBuffer.type_name()}")



import os

import numpy as np
import torch

from hive.agents.agent import Agent
from hive.utils.utils import seeder


class RandomAgent(Agent):
    """An agent that takes random steps at each timestep."""

    def __init__(self, obs_dim, act_dim, id=0, logger=None):
        """
        Args:
            obs_dim: The shape of the observations.
            act_dim (int): The number of actions available to the agent.
            id: Agent identifier.
            logger (ScheduledLogger): Logger used to log agent's metrics.
        """
        super().__init__(obs_dim=obs_dim, act_dim=act_dim, id=id)
        self._rng = np.random.default_rng(seed=seeder.get_new_seed())

    @torch.no_grad()
    def act(self, observation):
        """Returns a random action for the agent."""

        action = self._rng.integers(self._act_dim)

        return action

    def update(self, update_info):

        pass

    def save(self, dname):
        torch.save(
            {
                "rng": self._rng,
            },
            os.path.join(dname, "agent.pt"),
        )

    def load(self, dname):
        checkpoint = torch.load(os.path.join(dname, "agent.pt"))
        self._rng = checkpoint["rng"]



from hive.agents.qnets.conv import ConvNetwork


class NatureAtariDQNModel(ConvNetwork):
    """The convolutional network used to train the DQN agent in the original
    Nature paper: https://www.nature.com/articles/nature14236
    """

    def __init__(self, in_dim):
        """
        Args:
            in_dim (tuple): The tuple of observations dimension
                (channels, width, height).
        """
        super().__init__(
            in_dim=in_dim,
            channels=[32, 64, 64],
            kernel_sizes=[8, 4, 3],
            strides=[4, 2, 1],
            paddings=[0, 1, 1],
            mlp_layers=[512],
        )



class EnvSpec:
    """Object used to store information about environment configuration.
    Every environment should create an EnvSpec object.
    """

    def __init__(self, env_name, obs_dim, act_dim, env_info=None):
        """
        Args:
            env_name: Name of the environment
            obs_dim: Dimensionality of observations from environment. This can
                be a simple integer, or a complex object depending on the types
                of observations expected.
            act_dim: Dimensionality of action space.
            env_info: Any other info relevant to this environment. This can
                include items such as random seeds or parameters used to create
                the environment
        """
        self._env_name = env_name
        self._obs_dim = obs_dim
        self._act_dim = act_dim
        self._env_info = {} if env_info is None else env_info

    @property
    def env_name(self):
        return self._env_name

    @property
    def obs_dim(self):
        return self._obs_dim

    @property
    def act_dim(self):
        return self._act_dim

    @property
    def env_info(self):
        return self._env_info



import os
import pickle

import numpy as np

from hive.replays.replay_buffer import BaseReplayBuffer
from hive.utils.utils import create_folder, seeder


class CircularReplayBuffer(BaseReplayBuffer):
    """An efficient version of a circular replay buffer that only stores each observation
    once.
    """

    def __init__(
        self,
        capacity: int = 10000,
        stack_size: int = 1,
        n_step: int = 1,
        gamma: float = 0.99,
        observation_shape=(),
        observation_dtype=np.uint8,
        action_shape=(),
        action_dtype=np.int8,
        reward_shape=(),
        reward_dtype=np.float32,
        extra_storage_types=None,
        num_players_sharing_buffer: int = None,
    ):
        """Constructor for CircularReplayBuffer.

        Args:
            capacity (int): Total number of observations that can be stored in the
                buffer. Note, this is not the same as the number of transitions that
                can be stored in the buffer.
            stack_size (int): The number of frames to stack to create an observation.
            n_step (int): Horizon used to compute n-step return reward
            gamma (float): Discounting factor used to compute n-step return reward
            observation_shape: Shape of observations that will be stored in the buffer.
            observation_dtype: Type of observations that will be stored in the buffer.
                This can either be the type itself or string representation of the
                type. The type can be either a native python type or a numpy type. If
                a numpy type, a string of the form np.uint8 or numpy.uint8 is
                acceptable.
            action_shape: Shape of actions that will be stored in the buffer.
            action_dtype: Type of actions that will be stored in the buffer. Format is
                described in the description of observation_dtype.
            action_shape: Shape of actions that will be stored in the buffer.
            action_dtype: Type of actions that will be stored in the buffer. Format is
                described in the description of observation_dtype.
            reward_shape: Shape of rewards that will be stored in the buffer.
            reward_dtype: Type of rewards that will be stored in the buffer. Format is
                described in the description of observation_dtype.
            extra_storage_types (dict): A dictionary describing extra items to store
                in the buffer. The mapping should be from the name of the item to a
                (type, shape) tuple.
            num_players_sharing_buffer (int): Number of agents that share their
                buffers. It is used for self-play.
        """
        self._capacity = capacity
        self._specs = {
            "observation": (observation_dtype, observation_shape),
            "done": (np.uint8, ()),
            "action": (action_dtype, action_shape),
            "reward": (reward_dtype, reward_shape),
        }
        if extra_storage_types is not None:
            self._specs.update(extra_storage_types)
        self._storage = self._create_storage(capacity, self._specs)
        self._stack_size = stack_size
        self._n_step = n_step
        self._gamma = gamma
        self._discount = np.asarray(
            [self._gamma ** i for i in range(self._n_step)],
            dtype=self._specs["reward"][0],
        )
        self._episode_start = True
        self._cursor = 0
        self._num_added = 0
        self._rng = np.random.default_rng(seed=seeder.get_new_seed())
        self._num_players_sharing_buffer = num_players_sharing_buffer
        if num_players_sharing_buffer is not None:
            self._episode_storage = [[] for _ in range(num_players_sharing_buffer)]

    def size(self):
        """Returns the number of transitions stored in the buffer."""
        return max(
            min(self._num_added, self._capacity) - self._stack_size - self._n_step + 1,
            0,
        )

    def _create_storage(self, capacity, specs):
        """Creates the storage buffer for each type of item in the buffer.
        Args:
            capacity: The capacity of the buffer.
            specs: A dictionary mapping item name to a tuple (type, shape) describing
                the items to be stored in the buffer.
        """
        storage = {}
        for key in specs:
            dtype, shape = specs[key]
            dtype = str_to_dtype(dtype)
            specs[key] = dtype, shape
            shape = (capacity,) + tuple(shape)
            storage[key] = np.zeros(shape, dtype=dtype)
        return storage

    def _add_transition(self, **transition):
        """Internal method to add a transition to the buffer."""
        for key in transition:
            if key in self._storage:
                self._storage[key][self._cursor] = transition[key]
        self._num_added += 1
        self._cursor = (self._cursor + 1) % self._capacity

    def _pad_buffer(self, pad_length):
        """Adds padding to the buffer. Used when stack_size > 1, and padding needs to
        be added to the beginning of the episode.
        """
        for _ in range(pad_length):
            transition = {
                key: np.zeros_like(self._storage[key][0]) for key in self._storage
            }
            self._add_transition(**transition)

    def add(self, observation, action, reward, done, **kwargs):
        """Adds a transition to the buffer.
        The required components of a transition are given as positional arguments. The
        user can pass additional components to store in the buffer as kwargs as long as
        they were defined in the specification in the constructor.
        """

        if self._episode_start:
            self._pad_buffer(self._stack_size - 1)
            self._episode_start = False
        transition = {
            "observation": observation,
            "action": action,
            "reward": reward,
            "done": done,
        }
        transition.update(kwargs)
        for key in self._specs:
            obj_type = (
                transition[key].dtype
                if hasattr(transition[key], "dtype")
                else type(transition[key])
            )
            if not np.can_cast(obj_type, self._specs[key][0], casting="same_kind"):
                raise ValueError(
                    f"Key {key} has wrong dtype. Expected {self._specs[key][0]},"
                    f"received {type(transition[key])}."
                )
        if self._num_players_sharing_buffer is None:
            self._add_transition(**transition)
        else:
            self._episode_storage[kwargs["agent_id"]].append(transition)
            if done:
                for transition in self._episode_storage[kwargs["agent_id"]]:
                    self._add_transition(**transition)
                self._episode_storage[kwargs["agent_id"]] = []

        if done:
            self._episode_start = True

    def _get_from_array(self, array, indices, num_to_access=1):
        """Retrieves consecutive elements in the array, wrapping around if necessary.
        If more than 1 element is being accessed, the elements are concatenated along
        the first dimension.
        Args:
            array: array to access from
            indices: starts of ranges to access from
            num_to_access: how many consecutive elements to access
        """
        full_indices = np.indices((indices.shape[0], num_to_access))[1]
        full_indices = (full_indices + np.expand_dims(indices, axis=1)) % (
            self.size() + self._stack_size + self._n_step - 1
        )
        elements = array[full_indices]
        elements = elements.reshape(indices.shape[0], -1, *elements.shape[3:])
        return elements

    def _get_from_storage(self, key, indices, num_to_access=1):
        """Gets values from storage.
        Args:
            key: The name of the component to retrieve.
            indices: This can be a single int or a 1D numpyp array. The indices are
                adjusted to fall within the current bounds of the buffer.
            num_to_access: how many consecutive elements to access
        """
        if not isinstance(indices, np.ndarray):
            indices = np.array([indices])
        if num_to_access == 0:
            return np.array([])
        elif num_to_access == 1:
            return self._storage[key][
                indices % (self.size() + self._stack_size + self._n_step - 1)
            ]
        else:
            return self._get_from_array(
                self._storage[key], indices, num_to_access=num_to_access
            )

    def _sample_indices(self, batch_size):
        """Samples valid indices that can be used by the replay."""
        indices = np.array([], dtype=np.int32)
        while len(indices) < batch_size:
            start_index = (
                self._rng.integers(self.size(), size=batch_size - len(indices))
                + self._cursor
            )
            start_index = self._filter_transitions(start_index)
            indices = np.concatenate([indices, start_index])
        return indices + self._stack_size - 1

    def _filter_transitions(self, indices):
        """Filters invalid indices."""
        if self._stack_size == 1:
            return indices
        done = self._get_from_storage("done", indices, self._stack_size - 1)
        done = done.astype(bool)
        if self._stack_size == 2:
            indices = indices[~done]
        else:
            indices = indices[~done.any(axis=1)]
        return indices

    def sample(self, batch_size):
        """Sample transitions from the buffer. For a given transition, if it's
        done is True, the next_observation value should not be taken to have any
        meaning.

        Args:
            batch_size (int): Number of transitions to sample.
        """
        if self._num_added < self._stack_size + self._n_step:
            raise ValueError("Not enough transitions added to the buffer to sample")
        indices = self._sample_indices(batch_size)
        batch = {}
        batch["indices"] = indices
        terminals = self._get_from_storage("done", indices, self._n_step)

        if self._n_step == 1:
            is_terminal = terminals
            trajectory_lengths = np.ones(batch_size)
        else:
            is_terminal = terminals.any(axis=1).astype(int)
            trajectory_lengths = (
                np.argmax(terminals.astype(bool), axis=1) + 1
            ) * is_terminal + self._n_step * (1 - is_terminal)
        trajectory_lengths = trajectory_lengths.astype(np.int64)

        for key in self._specs:
            if key == "observation":
                batch[key] = self._get_from_storage(
                    "observation",
                    indices - self._stack_size + 1,
                    num_to_access=self._stack_size,
                )
            elif key == "done":
                batch["done"] = is_terminal
            elif key == "reward":
                rewards = self._get_from_storage("reward", indices, self._n_step)
                if self._n_step == 1:
                    rewards = np.expand_dims(rewards, 1)
                rewards = rewards * np.expand_dims(self._discount, axis=0)

                # Mask out rewards past trajectory length
                mask = np.expand_dims(trajectory_lengths, 1) > np.arange(self._n_step)
                rewards = np.sum(rewards * mask, axis=1)
                batch["reward"] = rewards
            else:
                batch[key] = self._get_from_storage(key, indices)

        batch["trajectory_lengths"] = trajectory_lengths
        batch["next_observation"] = self._get_from_storage(
            "observation",
            indices + trajectory_lengths - self._stack_size + 1,
            num_to_access=self._stack_size,
        )
        return batch

    def save(self, dname):
        """Save the replay buffer.

        Args:
            dname (str): directory where to save buffer. Should already have been
                created.
        """
        storage_path = os.path.join(dname, "storage")
        create_folder(storage_path)
        for key in self._specs:
            np.save(
                os.path.join(storage_path, f"{key}"),
                self._storage[key],
                allow_pickle=False,
            )
        state = {
            "episode_start": self._episode_start,
            "cursor": self._cursor,
            "num_added": self._num_added,
            "rng": self._rng,
        }
        with open(os.path.join(dname, "replay.pkl"), "wb") as f:
            pickle.dump(state, f)

    def load(self, dname):
        """Load the replay buffer.

        Args:
            dname (str): directory where to load buffer from.
        """
        storage_path = os.path.join(dname, "storage")
        for key in self._specs:
            self._storage[key] = np.load(
                os.path.join(storage_path, f"{key}.npy"), allow_pickle=False
            )
        with open(os.path.join(dname, "replay.pkl"), "rb") as f:
            state = pickle.load(f)
        self._episode_start = state["episode_start"]
        self._cursor = state["cursor"]
        self._num_added = state["num_added"]
        self._rng = state["rng"]


class SimpleReplayBuffer(BaseReplayBuffer):
    """A simple circular replay buffers.

    Args:
            capacity (int): repaly buffer capacity
            compress (bool): if False, convert data to float32 otherwise keep it as
                int8.
            seed (int): Seed for a pseudo-random number generator.
    """

    def __init__(self, capacity=1e5, compress=False, seed=42, **kwargs):

        self._numpy_rng = np.random.default_rng(seed)
        self._capacity = int(capacity)
        self._compress = compress

        self._dtype = {
            "observation": "int8" if self._compress else "float32",
            "action": "int8",
            "reward": "int8" if self._compress else "float32",
            "next_observation": "int8" if self._compress else "float32",
            "done": "int8" if self._compress else "float32",
        }

        self._data = {}
        for data_key in self._dtype:
            self._data[data_key] = [None] * int(capacity)

        self._write_index = -1
        self._n = 0
        self._previous_transition = None

    def add(self, observation, action, reward, done, **kwargs):
        """
        Adds transition to the buffer

        Args:
            observation: The current observation
            action: The action taken on the current observation
            reward: The reward from taking action at current observation
            done: If current observation was the last observation in the episode
        """
        if self._previous_transition is not None:
            self._previous_transition["next_observation"] = observation
            self._write_index = (self._write_index + 1) % self._capacity
            self._n = int(min(self._capacity, self._n + 1))
            for key in self._data:
                self._data[key][self._write_index] = np.asarray(
                    self._previous_transition[key], dtype=self._dtype[key]
                )
        self._previous_transition = {
            "observation": observation,
            "action": action,
            "reward": reward,
            "done": done,
        }

    def sample(self, batch_size=32):
        """
        sample a minibatch

        Args:
            batch_size (int): The number of examples to sample.
        """
        if self.size() == 0:
            raise ValueError("Buffer does not have any transitions yet." % batch_size)

        indices = self._numpy_rng.integers(self._n, size=batch_size)
        rval = {}
        for key in self._data:
            rval[key] = np.asarray(
                [self._data[key][idx] for idx in indices], dtype="float32"
            )

        return rval

    def size(self):
        """
        returns the number of transitions stored in the replay buffer
        """
        return self._n

    def save(self, dname):
        """
        Saves buffer checkpointing information to file for future loading.

        Args:
            dname (str): directory name where agent should save all relevant info.
        """
        create_folder(dname)

        sdict = {}
        sdict["capacity"] = self._capacity
        sdict["write_index"] = self._write_index
        sdict["n"] = self._n
        sdict["data"] = self._data

        full_name = os.path.join(dname, "meta.ckpt")
        with open(full_name, "wb") as f:
            pickle.dump(sdict, f)

    def load(self, dname):
        """
        Loads buffer from file.

        Args:
            dname (str): directory name where buffer checkpoint info is stored.

        Returns:
            True if successfully loaded the buffer. False otherwise.
        """
        full_name = os.path.join(dname, "meta.ckpt")
        with open(full_name, "rb") as f:
            sdict = pickle.load(f)

        self._capacity = sdict["capacity"]
        self._write_index = sdict["write_index"]
        self._n = sdict["n"]
        self._data = sdict["data"]


def str_to_dtype(dtype):
    if isinstance(dtype, type):
        return dtype
    elif dtype.startswith("np.") or dtype.startswith("numpy."):
        return np.typeDict[dtype.split(".")[1]]
    else:
        type_dict = {
            "int": int,
            "float": float,
            "str": str,
            "bool": bool,
        }
        return type_dict[dtype]
